import os
import re
import sys
import time
import signal
import warnings

from . import timeout

__all__ = ["daemon", "daemonize"]


def daemon(stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
    """
    将进程变成守护进程
    :param stdin:
    :param stdout:
    :param stderr:
    :return:
    """
    # 重定向标准文件描述符（默认情况下定向到/dev/null）
    try:
        pid = os.fork()
        # 父进程(会话组头领进程)退出，这意味着一个非会话组头领进程永远不能重新获得控制终端。
        if pid > 0:
            # 父进程退出
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("fork #1 failed: (%d) %s\n" % (e.errno, e.strerror))
        sys.exit(1)

    # 从母体环境脱离
    # chdir确认进程不保持任何目录于使用状态，否则不能umount一个文件系统。
    # 也可以改变到对于守护程序运行重要的文件所在目录
    os.chdir("/")
    # 调用umask(0)以便拥有对于写的任何东西的完全控制，因为有时不知道继承了什么样的umask。
    os.umask(0)
    # setsid调用成功后，进程成为新的会话组长和新的进程组长，并与原来的登录会话和进程组脱离。
    os.setsid()

    # 执行第二次fork
    try:
        pid = os.fork()
        if pid > 0:
            # 第二个父进程退出
            sys.exit(0)
    except OSError as e:
        sys.stderr.write("fork #2 failed: (%d) %s\n" % (e.errno, e.strerror))
        sys.exit(1)

    # 进程已经是守护进程了，重定向标准文件描述符
    for f in sys.stdout, sys.stderr:
        f.flush()
    si = open(stdin, 'r')
    so = open(stdout, 'a+')
    se = open(stderr, 'a+')
    # dup2函数原子化关闭和复制文件描述符
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())


@timeout(5, "n")
def _t_raw_input(alive_pid):
    """重复启动时提示是否重启，5秒钟后超时返回n"""
    raw_input = globals().get("raw_input") or input
    return raw_input(
        "The process with pid %s is running, restart it? y/n: " % alive_pid)


def common_stop_start_control(
        parser, monitor_log_path, wait=2, allow_multi=False):
    warnings.warn(
        "common_stop_start_control is a deprecated alias,"
        " use daemonize instead.",
        DeprecationWarning, 2)
    return daemonize(parser, monitor_log_path, wait, allow_multi)


def daemonize(parser, monitor_log_path, wait=2, allow_multi=False):
    """
    开启关闭及守护进程方式启动的公共实现
    :param parser: 一个argparse命令行参数解析对象，其它参数请调用些函数之前声明
    :param monitor_log_path: 如果通过--daemon 守护进程的方式启动，
    需要提供守护进程的stdout, stderr的文件路径
    :param wait: 轮循发起关闭信号的间隔时间
    :param allow_multi: 是否允许开启多个进程
    :return: argparse的解析结果对象
    """
    parser.add_argument("-d", "--daemon", action="store_true", default=False)
    parser.add_argument(
        "method", nargs="?", choices=["stop", "start", "restart", "status"])
    args = parser.parse_args()
    pid = os.getpid()
    filter_process_name = "%s .*%s"%(sys.argv[0], "start")
    if args.method == "stop":
        _stop(filter_process_name, sys.argv[0], wait)
        sys.exit(0)
    if args.method == "status":
        alive_pid = _check_status(filter_process_name, [pid])
        prompt = ["PROCESS", "STATUS", "PID", "TIME"]
        status = [sys.argv[0].rstrip(".py"),
                  "RUNNING" if alive_pid else "STOPPED",
                  str(alive_pid), time.strftime("%Y-%m-%d %H:%M:%S")]
        for line in _format_line([prompt, status]):
            [print(i, end="") for i in line]
            print("")
        sys.exit(0)
    elif args.method == "restart":
        _stop(filter_process_name, sys.argv[0], wait, ignore_pid=[pid])
        print("Start new process. ")
    else:
        alive_pid = _check_status(filter_process_name, ignore_pid=[pid])
        if alive_pid and not allow_multi:
            result = _t_raw_input(alive_pid)
            if result.lower() in ["y", "yes"]:
                _stop(filter_process_name, sys.argv[0], wait, ignore_pid=[pid])
            else:
                sys.exit(0)
    if args.daemon:
        daemon(stdout=monitor_log_path, stderr=monitor_log_path)
    return args


def _format_line(lines):
    if not lines:
        return [""]
    max_lengths = [0]*len(lines[0])
    for line in lines:
        for i, ele in enumerate(line):
            max_lengths[i] = max(len(ele), max_lengths[i])
    new_lines = []
    for line in lines:
        new_line = []
        for i, ele in enumerate(line):
            new_line.append(ele.ljust(max_lengths[i]+1))
        new_lines.append(new_line)
    return new_lines


def _stop(name, default_name=None, timedelta=2, ignore_pid=[]):
    """
    关闭符合名字（regex)要求的进程
    :param name: regex用来找到相关进程
    :param default_name: 仅用做显示该进程的名字
    :param timedelta: 轮循发起关闭信号的间隔时间
    :param ignore_pid: 忽略关闭的进程号
    :return:
    """
    pid = _check_status(name, ignore_pid)
    if not pid:
        print("No such process named %s" % (default_name or name))
    while pid:
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError:
            pass
        pid = _check_status(name, ignore_pid)
        if pid:
            print("Wait for %s exit. pid: %s" % (default_name or name, pid))
        time.sleep(timedelta)


def _check_status(name, ignore_pid):
    """
    找到主进程的pid
    :param name:
    :param ignore_pid:
    :return:
    """
    ignore = re.compile(r'ps -ef|grep')
    check_process = os.popen('ps -ef|grep "%s"' % name)
    processes = [x for x in check_process.readlines() if x.strip()]
    check_process.close()
    main_pid, main_ppid = None, None
    for process in processes:
        col = process.split()
        pid, ppid = col[1: 3]
        pid = int(pid) if int(pid) not in ignore_pid else None
        ppid = int(ppid)
        if not pid or ignore.search(process):
            continue
        main_pid, main_ppid = (pid, ppid) if not main_pid else \
            ((pid, ppid) if main_ppid == pid else (main_pid, main_ppid))
    return main_pid and (int(main_pid)
                         if int(main_pid) not in ignore_pid else None)




import os

from markdown import markdown

from .monitors import Service


class MarkDownRender(object):
    tpl = """
    <html>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="%s">
    <style>
        .markdown-body {
            box-sizing: border-box;
            min-width: 200px;
            max-width: 980px;
            margin: 0 auto;
            padding: 45px;
        }

        @media (max-width: 767px) {
            .markdown-body {
                padding: 15px;
            }
        }
    </style> 
    <article class="markdown-body">
        %s
    </article>
    </html>
    """

    def __init__(self, css, file_path):
        self.css = css
        self.file_path = file_path

    def __enter__(self):
        return self

    @staticmethod
    def _walk(path):
        for root, dir, filenames in os.walk(path):
            for fn in filenames:
                if fn.endswith(".md"):
                    fp = os.path.join(root, fn)
                    # 返回文件名和文件目录深度，深度用来设置css相对路径
                    yield fp, fp.replace(path, "").strip("/").count("/")

    def render(self):
        """
        用来渲染file_path下的md，将其输出为html格式
        :return:
        """
        if os.path.isfile(self.file_path):
            files = [(self.file_path, 0)]
        else:
            files = self._walk(self.file_path)

        first_page = None
        for input, deep in files:
            buffer = markdown(open(input).read(),
                              extensions=['markdown.extensions.extra'])
            output = input + ".html"
            first_page = first_page or output

            with open(output, "w") as f:
                f.write(self.tpl % ("../" * deep + self.css, buffer))
        return first_page

    def render_css(self):
        if os.path.isfile(self.file_path):
            self.file_path = os.path.dirname(self.file_path)

        with open(os.path.join(self.file_path, self.css), "w") as f:
            f.write(open(os.path.join(os.path.dirname(__file__), self.css)).read())

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.render_css()


class MarkDownHelper(Service):
    """
    将markdown文件转换成html文件并打开
    """
    def run(self):
        mk_render = MarkDownRender(self.args.css, self.args.input)

        with mk_render:
            output = mk_render.render()

            if output:
                self.logger.debug(f"打开{output}.")
                os.system(f"open {output}")
            else:
                self.logger.info("未发现可转换的文件！")

    def enrich_parser_arguments(self):
        super(MarkDownHelper, self).enrich_parser_arguments()
        self.parser.add_argument("input", help="markdown file or path.")
        self.parser.add_argument(
            "-c", "--css", help="css file name.", default="github-markdown.css")


def main():
    MarkDownHelper().run()



from ..monitors import ProxyPool
from .translate_adapter import TranslateAdapter

__all__ = ["Translator"]


class Translator(ProxyPool, TranslateAdapter):
    """
        基于代理池的翻译类
    """
    name = "translator"
    web_site = None
    translate_timeout = None
    retry_times = None

    def __init__(self, settings):
        super(Translator, self).__init__(settings=settings)
        self.web_site = self.settings.get(
            "WEBSITE", "baidu,qq,google").split(",")
        self.retry_times = self.settings.get("TRANSLATE_RETRY_TIMES", 10)
        self.translate_timeout = self.settings.get("TRANSLATE_TIMEOUT", 10)
        self.headers = self.settings.get("HEADERS") or self.headers



import re
import ast

from abc import abstractmethod, ABC

from .. import unsigned_right_shift, shift_left_for_js, shift_right_for_js

__all__ = ["BaiduAcquirer", "GoogleAcquirer", "BingAcquirer", "QqAcquirer"]


class TokenAcquirer(ABC):
    kwargs = None

    @property
    @abstractmethod
    def host(self):
        """提供host地址来获取必要信息"""
        pass

    def __init__(self, session, headers, proxies):
        """
        初始化acquirer信息
        :param session:
        :param headers:
        :param proxies:
        """
        self.session = session
        self.headers = headers
        self.proxies = proxies
        self.key = None

    def update(self):
        """
        更新认证信息，使用resp调用auth方法
        :return:
        """
        self.auth(self.session.get(
            self.host, proxies=self.proxies, headers=self.headers, timeout=3))

    @abstractmethod
    def auth(self, resp):
        """
        可以从resp获取必要信息，将必要信息字段存储在self中。
        :param resp:
        :return:
        """
        pass

    def enrich(self, kwargs):
        """
        每次生成acquirer之后，调用enrich方法，将之前获取过的必要信息放回
        :param kwargs:
        :return:
        """
        self.kwargs = kwargs
        if kwargs:
            cookies = kwargs.pop("cookies", None)
            if cookies:
                self.session.cookies = cookies
            self.__dict__.update(kwargs)

    def adjust(self, text):
        """
        百度,google调用acquire方法所需参数专用
        :param text:
        :return:
        """
        return text

    def __enter__(self):
        """
        每次使用acquirer之前，若不存在key，则主动发起更新
        :return:
        """
        if not self.key:
            self.update()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        每次退出acquirer之前，若发生异常，则主动发起更新。
        保存必要信息到kwargs。供下次生成acquirer使用。
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        if exc_type:
            self.update()
        self.kwargs["cookies"] = self.session.cookies
        self.kwargs["key"] = self.key
        return exc_type is None

    def acquire(self, text):
        """
        获取字段校验码。
        :param text:
        :return:
        """
        if not self.key:
            self.update()
        text = self.adjust(text)
        S = list()
        v = 0
        for ch in text:
            A = ord(ch)
            if 128 > A:
                S.append(A)
            elif 2048 > A:
                S.append(shift_right_for_js(A, 6) | 192)
            elif 55296 == (64512 & A) and v + 1 < len(text) \
                    and 56320 == (64512 & ord(text[v + 1])):
                A = 65536 + shift_left_for_js((1023 & A), 10) + \
                    (1023 & ord(text[v]))
                S.append(shift_right_for_js(A, 18) | 240)
                S.append(shift_right_for_js(A, 12) & 63 | 128)
                S.append(63 & A | 128)
            else:
                S.append(shift_right_for_js(A, 12) | 224)
                S.append(shift_right_for_js(A, 6) & 63 | 128)
                S.append(63 & A | 128)
        m, s = self.key.split(".")
        m = int(m)
        s = int(s)
        r = m
        for b in S:
            r = r + b
            r = self.n(r, '+-a^+6')
        r = self.n(r, '+-3^+b+-f')
        r ^= s
        if 0 > r:
            r = (2147483647 & r) + 2147483648

        r %= 1e6
        return "%s.%s" % (int(r), int(r) ^ m)

    @staticmethod
    def n(r, o):
        for t in range(0, len(o) - 2, 3):
            a = o[t + 2]
            a = ord(a) - 87 if a >= "a" else int(a)
            a = unsigned_right_shift(r, a) if "+" == o[t + 1] else r << a
            r = r + a & 4294967295 if "+" == o[t] else r ^ a
        return r


class BingAcquirer(TokenAcquirer):
    host = "https://cn.bing.com/translator/"

    def __init__(self, host, session, headers=None):
        super(BingAcquirer, self).__init__(host, session, headers)

    def auth(self, resp):
        self.key = "0" # 无用，占位

    def acquire(self, text):
        code = 0
        for i in text:
            i = ord(i)
            code = shift_left_for_js(code, 5) - code + i | 0
        return code


class BaiduAcquirer(TokenAcquirer):
    host = "http://fanyi.baidu.com/"

    def update(self):
        self.session.get(
            self.host, proxies=self.proxies, headers=self.headers, timeout=3)
        # 要连发两次才能用
        self.auth(self.session.get(
            self.host, proxies=self.proxies, headers=self.headers, timeout=3))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.kwargs["token"] = self.token
        return super(BaiduAcquirer, self).__exit__(exc_type, exc_val, exc_tb)

    def auth(self, resp):
        self.token = re.search(r"token: '(\w+)'", resp.text).group(1)
        self.key = re.search(r"window.gtk = '(.*?)'", resp.text).group(1)

    def adjust(self, text):
        if len(text) > 30:
            return text[0: 10] + \
                   text[len(text) // 2 - 5: len(text) // 2 + 5] + text[-10:]
        return super(BaiduAcquirer, self).adjust(text)


class QqAcquirer(TokenAcquirer):
    host = "http://fanyi.qq.com/"

    def auth(self, resp):
        self.key = "0" # 无用，占位
        self.headers = dict()
        self.headers["Origin"] = "http://fanyi.qq.com"
        self.headers["X-Requested-With"] = "XMLHttpRequest"
        self.headers["Cookie"] = ";".join(re.findall(r'document.cookie = "(.*)"', resp.text))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.kwargs["headers"] = self.headers
        return super(QqAcquirer, self).__exit__(exc_type, exc_val, exc_tb)

    def acquire(self, text):
        return NotImplemented


class GoogleAcquirer(TokenAcquirer):
    host = 'https://translate.google.cn'
    RE_TKK = re.compile(
        r'TKK=eval\(\'\(\(function\(\)\{(.+?)\}\)\(\)\)\'\);', re.DOTALL)

    def auth(self, resp):
        code = str(self.RE_TKK.search(resp.text).group(1)).replace('var ', '')
        code = code.encode().decode('unicode-escape')

        if code:
            tree = ast.parse(code)
            visit_return = False
            operator = '+'
            n, keys = 0, dict(a=0, b=0)
            for node in ast.walk(tree):
                if isinstance(node, ast.Assign):
                    name = node.targets[0].id
                    if name in keys:
                        if isinstance(node.value, ast.Num):
                            keys[name] = node.value.n
                        # the value can sometimes be negative
                        elif isinstance(node.value, ast.UnaryOp) and \
                                isinstance(node.value.op, ast.USub):
                            keys[name] = -node.value.operand.n
                elif isinstance(node, ast.Return):
                    # parameters should be set after this point
                    visit_return = True
                elif visit_return and isinstance(node, ast.Num):
                    n = node.n
                elif visit_return and n > 0:
                    # the default operator is '+' but implement some more for
                    # all possible scenarios
                    if isinstance(node, ast.Add):  # pragma: nocover
                        pass
                    elif isinstance(node, ast.Sub):  # pragma: nocover
                        operator = '-'
                    elif isinstance(node, ast.Mult):  # pragma: nocover
                        operator = '*'
                    elif isinstance(node, ast.Pow):  # pragma: nocover
                        operator = '**'
                    elif isinstance(node, ast.BitXor):  # pragma: nocover
                        operator = '^'
            # a safety way to avoid Exceptions
            clause = compile('{1}{0}{2}'.format(
                operator, keys['a'], keys['b']), '', 'eval')
            value = eval(clause, dict(__builtin__={}))
            self.key = '{}.{}'.format(n, value)



import re
import os
import sys
import json
import random
import logging
import requests
import traceback

from threading import local
from functools import partial
from collections import defaultdict
from abc import ABC, abstractmethod

from . import sites
from .. import retry_wrapper
from .token_acquirer import *

__all__ = ["TranslateAdapter"]


class TranslateAdapter(ABC):
    headers = {
        'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) '
                      'AppleWebKit/537.36 (KHTML, like Gecko)'
                      ' Chrome/41.0.2272.76 Safari/537.36',
        'Accept': 'text/html,application/xhtml+xml,application'
                  '/xml;q=0.9,*/*;q=0.8',
        "Accept-Language": "en-US,en;q=0.5",
    }
    local = local()

    def __init__(self):
        self.proxy = {}
        self.load(sites)

    @property
    @abstractmethod
    def web_site(self):
        pass

    @property
    @abstractmethod
    def translate_timeout(self):
        pass

    @property
    @abstractmethod
    def retry_times(self):
        pass

    def __enter__(self):
        self.session = requests.Session()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()

    def set_logger(self, logger=None):
        if not logger:
            self.logger = logging.getLogger()
            self.logger.setLevel(20)
            self.logger.addHandler(logging.StreamHandler(sys.stdout))
        else:
            self.logger = logger
            self.name = logger.name

    def load(self, module_str):
        if isinstance(module_str, str):
            sys.path.insert(0, os.getcwd())
            model = __import__(module_str, fromlist=module_str.split(".")[-1])
        else:
            model = module_str

        for k in dir(model):
            v = getattr(model, k)
            if hasattr(v, "__call__"):
                self.__dict__[k] = partial(v, self)

    @abstractmethod
    def proxy_choice(self):
        pass

    def trans_error_handler(self, func_name, retry_time, e, *args, **kwargs):
        """
        error_handler实现参数
        :param func_name: 重试函数的名字
        :param retry_time: 重试到了第几次
        :param e: 需要重试的异常
        :param args: 重试参数的参数
        :param kwargs: 重试参数的参数
        :return: 当返回True时，该异常不会计入重试次数
        """
        self.logger.debug("Error in %s for retry %s times. Error: %s" % (
            func_name, retry_time, e))
        proxies = self.proxy_choice()
        if proxies:
            args[1].update(proxies)

    def translate(self, src):
        """
        翻译主函数
        :param src: 源
        :return: 结果
        """
        try:
            # 找出大于号和小于号之间的字符，使用换行符连接，进行翻译
            pattern = re.compile(r"(?:^|(?<=>))([\s\S]*?)(?:(?=<)|$)")
            ls = re.findall(pattern, src.replace("\n", ""))
            src_data = "\n".join(x.strip("\t ") for x in ls if x.strip())
            if src_data.strip():
                # 对源中的%号进行转义
                src_escape = src.replace("%", "%%")
                # 将源中被抽离进行翻译的部分替换成`%s`，
                #  如果被抽离部分没有实质内容（为空），则省略
                src_template = re.sub(
                    pattern,
                    lambda x: "%s" if x.group(1).strip() else "", src_escape)
                return retry_wrapper(
                    self.retry_times,
                    error_handler=self.trans_error_handler)(
                    self._translate)(
                    src_data, self.proxy or self.proxy_choice()
                              or self.proxy, src_template)
        except Exception:
            self.logger.error(
                "Error in translate, finally, we could not get "
                "the translate result. src: %s, Error:  %s" % (
                    src, traceback.format_exc()))
        return src

    def _translate(self, src, proxies, src_template):
        web_site = random.choice(self.web_site).strip()
        acq = globals().get("%sAcquirer" % web_site.capitalize())
        if acq:
            if not getattr(self.local, "acquirers", None):
                self.local.acquirers = dict()
            if not getattr(self.local, "proxies_cookies", None):
                self.local.proxies_cookies = defaultdict(dict)
            if web_site not in self.local.acquirers:
                self.local.acquirers[web_site] = acq(
                    self.session, self.headers, proxies)
            self.local.acquirers[web_site].enrich(
                self.local.proxies_cookies.setdefault(
                    json.dumps(proxies), dict()))
            return getattr(self, web_site)(
                src, proxies, src_template, self.local.acquirers[web_site])
        else:
            return getattr(self, web_site)(src, proxies, src_template)



import os
import sys
import signal
import random
import logging
import warnings
import traceback

from redis import Redis
from itertools import repeat
from argparse import ArgumentParser
from kafka import KafkaConsumer, KafkaProducer

from .logger import Logger
from .frozen import Frozen
from .combine import combine
from .daemon import daemonize
from .consoler import Consoler
from .singleton import Singleton
from .managers import ExceptContext
from .settings import SettingsLoader

from . import call_later, cache_property

__all__ = ["ParallelMonitor", "LoggingMonitor", "Service",
           "ProxyPool", "ItemConsumer", "ItemProducer"]


class ParallelMonitor(object, metaclass=Singleton):
    """
    支持多线程多进程统一管理
    """
    alive = True
    name = "parallel_monitor"
    children = []
    int_signal_count = 1

    def __init__(self):
        self.open()
        super(ParallelMonitor, self).__init__()

    @cache_property
    def logger(self):
        logger = logging.getLogger(self.name)
        logger.setLevel(10)
        logger.addHandler(logging.StreamHandler(sys.stdout))
        return logger

    def stop(self, *args):
        if self.int_signal_count > 1:
            self.logger.info("Force to terminate...")
            for th in self.children[:]:
                self.stop_child(th)
            pid = os.getpid()
            os.kill(pid, 9)

        else:
            self.alive = False
            self.logger.info("Close process %s..." % self.name)
            self.int_signal_count += 1

    def open(self):
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)

    def stop_child(self, child):
        pass


class LoggingMonitor(object):
    """
    内建Logger和Settings
    """
    loader = SettingsLoader()
    default_settings = "settings"

    def __init__(self, settings=None, local_settings=None, name=None):
        if name:
            self.name = name
        if isinstance(settings, Frozen):
            self.settings = settings
        else:
            self.settings = self.loader.load(
                local=local_settings, default=settings or self.default_settings)
            super(LoggingMonitor, self).__init__()

    @cache_property
    def logger(self):
        return Logger(self.settings, self.name)

    def log_err(self, func_name, *args):
        self.logger.error("Error in %s: %s. " % (
            func_name, "".join(traceback.format_exception(*args))))
        return True


@combine(Consoler,
         ("args.console_host", "args.console_port", "args.console", "debug"))
@combine(LoggingMonitor, ("args.settings", "args.localsettings"))
class Service(ParallelMonitor):
    """
        可执行程序，支持守护进程启动
    """
    name = "Service"
    parser = None
    args = None

    def __init__(self):
        self.args = self.parse_args()
        super(Service, self).__init__()

    def enrich_parser_arguments(self):
        self.parser.add_argument("-s", "--settings", help="Setting module. ")
        self.parser.add_argument(
            "-ls", "--localsettings", help="Local setting module. ",
            default="localsettings")

    def parse_args(self, daemon_log_path='/dev/null'):
        self.parser = ArgumentParser(conflict_handler="resolve")
        self.enrich_parser_arguments()
        return daemonize(
            self.parser, daemon_log_path, 2,
            bool(eval(os.environ.get("DEBUG", "0").lower().capitalize())))

    @property
    def debug(self):
        return self.settings.get_bool("DEBUG", False)


class ProxyPool(LoggingMonitor):
    """
        Redis代理池
    """
    def __init__(self, settings):
        """
        proxy_sets指定多个redis set,随机选取proxy
        proxy
        :param settings:
        """
        super(ProxyPool, self).__init__(local_settings=settings)
        self.protocols = self.settings.get("PROTOCOLS", "http,https").split(",")
        self.redis_conn = Redis(
            self.settings.get("REDIS_HOST", "0.0.0.0"),
            self.settings.get_int("REDIS_PORT", 6379))
        self.proxy_sets = self.settings.get("PROXY_SETS", "proxy_set").split(",")
        self.account_password = self.settings.get("PROXY_ACCOUNT_PASSWORD", "")
        self.proxy = {}

    def proxy_choice(self):
        """
        随机选取代理
        :return: 代理
        """
        proxy = self.redis_conn.srandmember(random.choice(self.proxy_sets))
        if proxy:
            proxy_str = "http://%s%s" % (
                self.account_password+"@" if self.account_password else "",
                proxy.decode())
            self.proxy = dict(zip(self.protocols, repeat(proxy_str)))
        return self.proxy


class ItemConsumer(Service):
    """
    kafka consumer.
    """
    name = "item_consumer"

    def __init__(self):
        super(ItemConsumer, self).__init__()
        self.topic_in = self.args.topic_in
        self.consumer_id = self.args.consumer or self.name
        with ExceptContext(Exception, errback=lambda *args: self.stop()):
            self.consumer = KafkaConsumer(self.topic_in,
                                          group_id=self.consumer_id,
                                          bootstrap_servers=
                                          self.settings.KAFKA_HOSTS.split(","),
                                          enable_auto_commit=False,
                                          consumer_timeout_ms=1000)

    def consume_errback(self, exception):
        self.logger.error("Consume feature failed: %s. " % str(exception))

    def consume_callback(self, value):
        self.logger.debug("Consume feature success: %s. " % str(value))

    @call_later("commit", interval=10, immediately=False)
    def consume(self, callback=lambda value: value, errback=lambda message: None):
        """
        两种模式，当不提供callback, errback的时候，
        调用consume直接获得message.value，异步处理信息
        当提供了自定义callback,则处理信息变为同步，
        可以选择在callback中对message.value进行处理，返回假则处理失败，
        处理失败后会继续调用errback，可以做回滚offset等操作。
        :param callback:
        :param errback:
        :return:
        """
        with ExceptContext(errback=self.log_err):
            with ExceptContext(
                    (StopIteration, OSError), errback=lambda *args: True):
                message = self.consumer.__next__()
                while message:
                    value = message.value
                    self.logger.debug(
                        'Consume message from %s, message is %s' % (
                            self.topic_in, value))
                    if isinstance(value, bytes):
                        value = value.decode("utf-8")
                    value = callback(value)
                    if value:
                        return value
                    else:
                        return errback(message)

    def commit(self):
        future = self.consumer.commit_async()
        future.add_callback(self.consume_callback)
        future.add_errback(self.consume_errback)

    def close(self):
        with ExceptContext(Exception, errback=lambda *args: True):
            if self.consumer:
                self.consumer.close()

    def enrich_parser_arguments(self):
        super(ItemConsumer, self).enrich_parser_arguments()
        self.parser.add_argument("-ti", "--topic-in", help="Topic in. ")
        self.parser.add_argument("-c", "--consumer", help="consumer id. ")


class ItemProducer(Service):
    """
    kafka producer.
    """
    name = "item_producer"

    def __init__(self):
        super(ItemProducer, self).__init__()
        self.topic_out = self.args.topic_out
        with ExceptContext(Exception, errback=lambda *args: self.stop()):
            self.producer = KafkaProducer(
                bootstrap_servers=self.settings.KAFKA_HOSTS.split(","),
                retries=3)

    def produce_callback(self, value):
        self.logger.debug("Product future success: %s"%value)

    def produce_errback(self, value):
        self.logger.error("Product future failed: %s" % value)

    def produce(self, *messages):
        with ExceptContext(Exception, errback=self.log_err):
            for message in messages:
                if isinstance(message, str):
                    message = message.encode("utf-8")
                future = self.producer.send(self.topic_out, message)
                future.add_callback(self.produce_callback)
                future.add_errback(self.produce_errback)
                future.get(timeout=50)

    def enrich_parser_arguments(self):
        super(ItemProducer, self).enrich_parser_arguments()
        self.parser.add_argument("-to", "--topic-out", help="Topic out. ")



import json
import time


def merge_conflict(src_template, returns):
    return src_template % tuple(returns[:src_template.count("%s")]) + \
           "。".join(returns[src_template.count("%s"):])


def youdao(self, src_data, proxies, src_template):
    """
    有道翻译的实现(废弃)
    :param src_data: 原生数据
    :param proxies: 代理
    :param src_template: 原生数据模板
    :return: 结果
    """
    url = "http://fanyi.youdao.com/translate"
    resp = self.session.post(url=url, data={
        'keyfrom': 'fanyi.web',
        'i': src_data,
        'doctype': 'json',
        'action': 'FY_BY_CLICKBUTTON',
        'ue': 'UTF-8',
        'xmlVersion': '1.8',
        'type': 'AUTO',
        'typoResult': 'true'}, headers=self.headers,
                         timeout=self.translate_timeout, proxies=proxies)
    return src_template % tuple(map(lambda y: "".join(
        map(lambda x: x["tgt"], y)), json.loads(resp.text)["translateResult"]))


def bing(self, src_data, proxies, src_template, acquirer):
    """
    必应翻译的实现
    :param src_data: 原生数据
    :param proxies: 代理
    :param src_template: 原生数据模板
    :return: 结果
    """
    with acquirer:
        url = "https://cn.bing.com/translator/api/Translate/TranslateArray?" \
              "from=-&to=zh-CHS"
        data_mapping = dict(
            (d, acquirer.acquire(d)) for d in src_data.split("\n"))
        resp = self.session.post(
            url=url,
            json=[{"id": acquirer.acquire(d), "text": d}
                  for d in src_data.split("\n")],
            headers=self.headers,
            timeout=self.translate_timeout, proxies=proxies)
        trans_rs = dict((item["id"], item["text"])
                        for item in json.loads(resp.text)["items"])
        finish_data = [trans_rs[str(data_mapping.get(d))]
                       for d in src_data.split("\n")]
        return src_template % tuple(finish_data)


def baidu(self, src_data, proxies, src_template, acquirer):
    """
    百度翻译的实现, 百度翻译最长只能翻译5000个字符
    :param src_data: 原生数据
    :param proxies: 代理
    :param src_template: 原生数据模板
    :param acquirer: token获取器
    :return: 结果
    """
    with acquirer:
        url = "http://fanyi.baidu.com/v2transapi"
        resp = self.session.post(url=url, data={
            'from': 'en',
            'to': 'zh',
            'transtype': 'realtime',
            'query': src_data,
            'simple_means_flag': 3,
            'token': acquirer.token,
            "sign": acquirer.acquire(src_data)}, headers=self.headers,
                             timeout=self.translate_timeout, proxies=proxies)
        return src_template % tuple("".join(
            map(lambda x: x["src_str"],
                json.loads(resp.text)["trans_result"]['phonetic'])).split("\n"))


def qq(self, src_data, proxies, src_template, acquirer):
    """
    腾讯翻译的实现, 腾讯翻译最长只能翻译2000个字符
    :param src_data: 原生数据
    :param proxies: 代理
    :param src_template: 原生数据模板
    :param acquirer: cookie获取器
    :return: 结果
    """
    url = 'http://fanyi.qq.com/api/translate'
    with acquirer:
        acquirer.headers.update(self.headers)
        resp = self.session.post(
            url, data={'source': 'auto',
                       'target': 'en',
                       'sourceText': src_data,
                       'sessionUuid': 'translate_uuid%d' % (time.time()*1000)},
            headers=acquirer.headers, timeout=self.translate_timeout, proxies=proxies)
        return merge_conflict(
            src_template,
            [record["targetText"] for record in json.loads(
                resp.text)["translate"]["records"]
             if record.get("sourceText") != "\n"])


def google(self, src_data, proxies, src_template, acquirer):
    url = 'https://translate.google.cn/translate_a/single'
    with acquirer:
        data = {
            'client': 't',
            'sl': "auto",
            'tl': "zh",
            'hl': "zh",
            'dt': ['at', 'bd', 'ex', 'ld', 'md', 'qca', 'rw', 'rm', 'ss', 't'],
            'ie': 'UTF-8',
            'oe': 'UTF-8',
            'otf': 1,
            'ssel': 0,
            'tsel': 0,
            'tk': acquirer.acquire(src_data),
            'q': src_data,
        }
        resp = self.session.get(
            url, params=data, headers=self.headers,
            timeout=self.translate_timeout, proxies=proxies)
        return merge_conflict(
            src_template,
            [line[0] for line in json.loads(resp.text)[0] if line[0]])



# -*- coding:utf-8 -*-
import types

from functools import wraps
from operator import attrgetter


def combine(part, words=(), keywords=(), after=True, extend=False):
    """
      组合的装饰器实现，使用方法：
      @combine(Consoler)
      class Service():
          pass
      动态将Consoler组合到Service中，Service组合Consoler所有非下划线方法及属性(覆盖)，
      after参数确定有相同方法存在时调用的顺序，True时Consoler方法在Service后调用。
      extend参数确定是否是继承关系，这决定了在调用part方法时使用part对象还是被组合的对象
      对于__init__方法，Service先调用自己的__init__然后再调用Consoler的__init__方法,
      调用Consoler时传入的参数为words, keywords指定，必须为Service实例属性，如:
      ("args.console_host", "args.console_port"), keywords 与words参数一样，
      不过在传入__init__时使用关键字参数的形式。
      属性优先级：Service实例属性，Console类属性，Service类属性，Console实例属性。
      """
    def outer(source_cls):
        if not hasattr(source_cls, "combines"):
            setattr(source_cls, "combines", [part])
        else:
            source_cls.combines.append(part)
        if words:
            wds = attrgetter(*words)
        if keywords:
            kwds = attrgetter(*keywords)
        s_p = dir(source_cls)
        c_p = [prop for prop in dir(part) if not prop.startswith("_")]

        for p in c_p:
            if p in s_p:

                def closure():
                    prop1 = getattr(source_cls, p)
                    prop2 = getattr(part, p)
                    if isinstance(prop1, types.FunctionType) \
                            and isinstance(prop2, types.FunctionType):

                        @wraps(prop1)
                        def wrapper(*args, **kwargs):
                            if extend:
                                self = args[0]
                            else:
                                self = getattr(
                                    args[0], part.__name__.lower(), args[0])
                            if after:
                                result = prop1(*args, **kwargs)
                                prop2(self, *args[1:], **kwargs)
                                return result
                            else:
                                result = prop2(self, *args[1:], **kwargs)
                                prop1(*args, **kwargs)
                                return result
                        return wrapper
                    return prop2

                setattr(source_cls, p, closure())
            else:
                setattr(source_cls, p, getattr(part, p))

        __init = getattr(source_cls, "__init__",
                         gen_default_method("__init__", locals()))

        @wraps(__init)
        def _init(*args, **kwargs):
            __init(*args, **kwargs)
            if len(words) == 0:
                a = tuple()
            elif len(words) == 1:
                a = [wds(args[0])]
            else:
                a = wds(args[0])
            if len(keywords) == 0:
                k = {}
            elif len(keywords) == 1:
                k = {keywords[0]: kwds(args[0])}
            else:
                k = dict(zip(keywords, kwds(args[0])))
            parter = part(*a, **k)
            setattr(args[0], part.__name__.lower(), parter)
            setattr(parter, source_cls.__name__.lower(), args[0])

        setattr(source_cls, "__init__", _init)

        def closure():
            __getattr = getattr(source_cls, "__getattr__",
                                gen_default_method(
                                    "__getattr__", locals(),
                                    "raise AttributeError"))
            part_name = part.__name__.lower()

            @wraps(__getattr)
            def _getattr(*args, **kwargs):
                try:
                    result = __getattr(*args, **kwargs)
                    return result
                except AttributeError:
                    if args[1] in \
                            [c.__name__.lower() for c in source_cls.combines]:
                        raise
                    return getattr(
                        getattr(
                            args[0], part_name, args[0]), *args[1:], **kwargs)
            return _getattr

        setattr(source_cls, "__getattr__", closure())

        return source_cls

    return outer


def gen_default_method(name, namespace=locals(), option=""):
    def_temp = """def {name}(self, *args, **kwargs):
    if [b for b in source_cls.__bases__ if hasattr(b, "{name}")]:
        return super(source_cls, self).{name}(*args, **kwargs)
    {option}
    """.format(name=name, option=option)
    exec(def_temp, namespace)
    return eval(name, namespace)



# -*- coding:utf-8 -*-
import os
import glob
import json
import struct

from threading import RLock
from abc import ABC, abstractmethod

from . import thread_safe_for_method, call_later
from .managers import ExceptContext

__all__ = ["Queue", "RedisQueue", "FifoDiskQueue"]


class Queue(ABC):

    def __init__(self):
        self.lock = RLock()

    @abstractmethod
    def push(self, *strings):
        pass

    @abstractmethod
    def pop(self):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def rid(self, count):
        pass

    @abstractmethod
    def clear(self):
        pass


class RedisQueue(Queue):
    """
    redis队列
    """
    def __init__(self, redis_conn, key):
        super(RedisQueue, self).__init__()
        self.redis_conn = redis_conn
        self.key = key

    @thread_safe_for_method
    def push(self, *strings):
        self.redis_conn.rpush(self.key, *strings)

    @thread_safe_for_method
    def pop(self):
        return self.redis_conn.lpop(self.key)

    def __len__(self):
        return self.redis_conn.llen(self.key)

    @thread_safe_for_method
    def rid(self, count):
        pipe = self.redis_conn.pipeline()
        pipe.multi()
        pipe.lrange(self.key, 0, count-1).ltrim(self.key, count, -1)
        results, count = pipe.execute()
        return results

    def clear(self):
        self.redis_conn.delete(self.key)

    def close(self):
        pass


class FifoDiskQueue(Queue):
    """持久化 FIFO 队列."""

    szhdr_format = ">L"
    szhdr_size = struct.calcsize(szhdr_format)

    def __init__(self, path, chunk_size=100000):
        super(FifoDiskQueue, self).__init__()
        self.path = path
        if not os.path.exists(path):
            os.makedirs(path)
        self.info = self._load_info(chunk_size)
        self.chunk_size = self.info['chunk_size']
        self.headf = self._open_chunk(self.info['head'][0], 'ab+')
        self.tailf = self._open_chunk(self.info['tail'][0])
        os.lseek(self.tailf.fileno(), self.info['tail'][2], os.SEEK_SET)

    @thread_safe_for_method
    @call_later(callback="_save_info")
    def push(self, *strings):
        for string in strings:
            self._push(string)

    def _push(self, string):
        if not isinstance(string, bytes):
            raise TypeError('Unsupported type: {}'.format(type(string).__name__))
        hnum, hpos = self.info['head']
        hpos += 1
        szhdr = struct.pack(self.szhdr_format, len(string))
        os.write(self.headf.fileno(), szhdr + string)
        if hpos == self.chunk_size:
            hpos = 0
            hnum += 1
            self.headf.close()
            self.headf = self._open_chunk(hnum, 'ab+')
        self.info['size'] += 1
        self.info['head'] = [hnum, hpos]

    def _open_chunk(self, number, mode='rb'):
        return open(os.path.join(self.path, 'q%05d' % number), mode)

    @call_later(callback="_save_info")
    def rid(self, count):
        messages = []
        while len(messages) < count:
            message = self.pop()
            if message:
                messages.append(message)
            else:
                break
        return messages

    @thread_safe_for_method
    @call_later(callback="_save_info", immediately=False)
    def pop(self):
        tnum, tcnt, toffset = self.info['tail']
        if [tnum, tcnt] >= self.info['head']:
            return
        tfd = self.tailf.fileno()
        szhdr = os.read(tfd, self.szhdr_size)
        if not szhdr:
            return
        size, = struct.unpack(self.szhdr_format, szhdr)
        data = os.read(tfd, size)
        tcnt += 1
        toffset += self.szhdr_size + size
        if tcnt == self.chunk_size and tnum <= self.info['head'][0]:
            tcnt = toffset = 0
            tnum += 1
            self.tailf.close()
            os.remove(self.tailf.name)
            self.tailf = self._open_chunk(tnum)
        self.info['size'] -= 1
        self.info['tail'] = [tnum, tcnt, toffset]
        return data

    def clear(self):
        self._cleanup()
        self._load_info(self.chunk_size)

    def close(self):
        self.headf.close()
        self.tailf.close()
        self._save_info()
        if len(self) == 0:
            self._cleanup()

    def __len__(self):
        return self.info['size']

    def _load_info(self, chunk_size):
        infopath = self._info_path()
        if os.path.exists(infopath):
            with ExceptContext(json.decoder.JSONDecodeError):
                with open(infopath) as f:
                    return json.load(f)
            os.unlink(infopath)
        return {
                'chunk_size': chunk_size,
                'size': 0,
                'tail': [0, 0, 0],
                'head': [0, 0],
            }

    def _save_info(self):
        with open(self._info_path(), 'w') as f:
            json.dump(self.info, f)

    def _info_path(self):
        return os.path.join(self.path, 'info.json')

    def _cleanup(self):
        for x in glob.glob(os.path.join(self.path, 'q*')):
            os.remove(x)
        os.remove(os.path.join(self.path, 'info.json'))
        if not os.listdir(self.path):
            os.rmdir(self.path)


from . import _assert


class Node(object):
    """
    链表的节点
    """
    def __init__(self, value, left=None, right=None):
        self.value = value
        self.left = left
        self.right = right

    def __str__(self):
        return str(self.value)

    __repr__ = __str__


class Iterator(object):
    """
    迭代器的实现
    """
    def __init__(self, current_node, towards="right"):
        self.current_node = current_node
        self.towards = towards

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_node:
            value = self.current_node.value
        else:
            raise StopIteration()
        self.current_node = getattr(self.current_node, self.towards)
        return value


class LinkedList(object):
    """
    链表
    """
    def __init__(self, iterable):
        self.head = None
        self.tail = None
        self._size = 0
        last_node = self._iadd(iterable)

        if last_node is not None:
            self.tail = last_node

    def _iadd(self, iterable, last_node=None):
        """
        自增逻辑
        :param iterable: 被加的可迭代对象
        :param last_node:
        :return:
        """
        for val in iterable:
            last_node = Node(val, left=last_node)
            if last_node is None or last_node.left is None:
                self.head = last_node
            else:
                last_node.left.right = last_node
            self._size += 1
        return last_node

    def __len__(self):
        return self._size

    def __iter__(self):
        return Iterator(self.head)

    def insert(self, index, x):
        self._slice_setter(index, index, x)

    def reverse_iter(self):
        """
        反向迭代
        :return:
        """
        return Iterator(self.tail, "left")

    def __eq__(self, other):
        """
        判断两个对象是否相等
        :param other:
        :return:
        """
        if not isinstance(other, self.__class__):
            return False
        if self._size != len(other):
            return False

        for i, j in zip(self, other):
            if i != j:
                return False
        return True

    def __hash__(self):
        raise ValueError("Cannot hash!")

    def _adjust_index(self, index):
        """
        调整下标，统一成正数的形式，对于不合法的下标抛出异常
        :param index:
        :return:
        """
        if index >= self._size:
            raise IndexError(f"Index: {index}")
        else:
            now = (self._size + index)
            if now < 0:
                raise IndexError(f"Index: {index}")
            return now % self._size

    def __getitem__(self, key):
        if not isinstance(key, slice):
            return self._get_node(key).value
        else:
            return self.__class__(self._slice_parser(key))

    def _slice_setter(self, left, right, *sequence):
        """
        切片赋值逻辑
        :param left:
        :param right:
        :param sequence:
        :return:
        """
        # 先获取左边的节点，如果越界，则从整个链表右侧追加元素，直接返回
        try:
            left_node = self._get_node(left)
        except IndexError:
            self._iadd(sequence, self.tail)
            return
        # 如果右边的下标大于左边的，则认定为列表插入操作，直接将左右视为相同
        # 否则，获取右边的节点，若越界，则将右边节点置为None
        if right <= left:
            right_node = left_node
        else:
            try:
                right_node = self._get_node(right)
            except IndexError:
                right_node = None
        # 求左右的距离，这个距离内的节点会被覆盖掉
        distance = self._distance_between(left_node, right_node)
        # 对链表进行自增操作，从左边节点开始自增
        last_node = self._iadd(sequence, left_node.left)
        # 增加完毕后，记得将左边节点的引用置为空
        left_node.left = None
        # 对于右边结点为空的情况，此时右边部分已经全被覆盖了，所以tail需要换掉
        if right_node is None:
            self.tail = last_node
        # 否则，将右边节点和自增完的最后一个节点连接。
        else:
            last_node.right = right_node
            right_node.left = last_node
        # 最后调整链表的尺寸
        self._size -= distance

    @staticmethod
    def _distance_between(left, right):
        """
        求两个结点之间的距离
        :param left: 左边的node， 不能为null
        :param right: 右边的node，为null时为左边到最后的距离
        :return:
        """
        length = 0
        while left != right:
            left = left.right
            length += 1
        return length

    def __setitem__(self, key, value):
        if not isinstance(key, slice):
            self._get_node(key).value = value
        else:
            self._slice_setter(key.start or 0, key.stop or self._size, *value)

    def __iadd__(self, other):
        last_node = self._iadd(other, self.tail)
        if last_node is not None:
            self.tail = last_node
        return self

    def __add__(self, other):
        lst = self.__class__(self)
        lst += other
        return lst

    def __radd__(self, other):
        lst = self.__class__(other)
        lst += self
        return lst

    def _slice_parser(self, sl):
        """
        切片获取逻辑
        :param sl:
        :return:
        """
        # 对于未指定开始的，认为是从-1开始，注意，-1不是下标，因为-1的下标其实是_size-1
        start = sl.start or -1
        # 对于未指定结束的，则按结束为size，size同样也不是下标
        # 可以理解默认的start和stop都是链表两个下标的自减1和自加1，为什么这么设计呢，
        # 因为没写时候，该端点上的值是需要被完全迭代的，所以适当扩大该端点的范围是被请允许的。
        stop = sl.stop or self._size
        step = sl.step or 1
        # begin是开始的下标，对于递增的切片，begin应该是指定的开始和0的最大值，
        # 此时between的判断逻辑左闭右开。对于递减的切片，begin应该是指定的结束和
        # size-1的最小值，此时between的逻辑应该是左开右闭。

        if step > 0:
            begin = max(start, 0)
            between = self._get_between(True)
            iterator = zip(range(0, self._size), iter(self))
        else:
            begin = min(stop, self._size-1)
            between = self._get_between(False)
            iterator = zip(range(self._size-1, -1, -1), self.reverse_iter())

        # 迭代链表，其下标与链表本身同样递增或递减
        for i, val in iterator:
            # 如果下标在开始和结束之间，同时从开始计算步长合适，则yield值
            if between(i, start, stop) and (i - begin) % step == 0:
                yield val

    @staticmethod
    def _get_between(with_left=True):
        if with_left:
            return lambda index, left, right: left <= index < right
        else:
            return lambda index, left, right: left < index <= right

    def _get_node(self, index):
        _assert(isinstance(index, int), TypeError("Index type must be int!"))
        index = self._adjust_index(index)
        node = self.head
        while index > 0:
            node = node.right
            index -= 1
        return node

    def _get_index_node_by_value(self, value, start=None, end=None):
        node = self.head
        index = 0
        if start is None:
            start = 0
        if end is None:
            end = self._size
        # 从头部开始查找等于目标值的节点，直到等于end时停止
        while index < end:
            if index >= start and node.value == value:
                return index, node

            node = node.right
            index += 1

        raise ValueError(f"Value: {value} not found!")

    def remove(self, value):
        _, node = self._get_index_node_by_value(value)
        return self._remove_node(node)

    def find(self, value, start=None, end=None):
        index, node = self._get_index_node_by_value(value, start, end)
        return index

    def pop(self, index=-1):
        node = self._get_node(index)
        self._remove_node(node)
        return node.value

    def _remove_node(self, node):
        # 如果该节点存在左边节点，则将该节点左边的节点指向该节点
        if node.left:
            node.left.right = node.right
        # 否则证明这是头部
        else:
            self.head = node.right

        # 同上
        if node.right:
            node.right.left = node.left
        else:
            self.tail = node.left

        self._size -= 1
        return True

    def __str__(self):
        if not self:
            return "[]"
        string = "["
        for i in self:
            string += "%r, " % i
        return string[:-2] + "]"

    __repr__ = __str__



# -*- coding:utf-8 -*-
from functools import reduce
from .async_context import contextmanager


class Processor(object):
    """
    进度管理器，管理收集指定区域业务逻辑流程中的进度执行情况，如现有一个函数执行某个功能，
    现需要掌握其执行进度，假设其整个流程可分为三步，分别是三个函数调用，其中进度权重大约是
    1:2:3，则创建一个processor，指定其weight为[1, 2, 3]，将processor传入该函数，并在
    该函数完成每步进行processor.update调用来更新进度，调用三次，则进度到达100%，
    如果每步的子函数需要进一步划分进度，则通过调用：

    .. code-block:: python

        with processor.hand_out(weight) as child_processor
            child_function(..., child_processor)

    child_processor将负责将该步的进度按权重分发到子函数中去。
    """

    def __init__(self, weight, update_callback, _from=0, to=100, skip=None):
        """

        :param weight: 将一个函数分成的工作区块数，如果是int类型，则是区作区块数，
            如果是[1, 4, 6...]类型，则指定了每个工作区块的工作量权重。
        :param update_callback: 当完成规定的进度后，调用的函数
        :param _from:
        :param to:
        :param skip:
        """
        assert _from <= to, "Process should be increase! Got: %s-%s" % (_from, to)
        if isinstance(weight, int):
            weight = [1] * weight

        self.processes = self.split(weight, _from, to)
        self.update_callback = update_callback
        self.last_process = _from
        self._skip = skip

    def can_update(self, current_process):
        return not (self._skip and self._skip(current_process))

    def skip(self):
        """
        通过判断下一个进度值是否小于当前进度来决定是否跳过
        @return:
        """
        return not self.can_update(self.processes[0])

    def skip_all(self):
        """
        如果派生了子进度条，则通过判断子进度最大进度值来决定是否直接跳过子进度条
        @return:
        """
        return not self.can_update(self.processes[-1])

    @staticmethod
    def split(weight, _from, to):
        # 没有权重时，直接返回
        if not weight:
            return weight

        count = sum(weight)
        distance = float(to - _from)
        step = distance / float(count)
        # 如果步长是0， 则证明每一步和开始的进度是一样的
        if step == 0:
            return [to] * len(weight)
        else:
            processes = []
            reduce(lambda w1, w2: processes.append(
                int(_from + step*(w1 + w2))) or w1 + w2, weight, 0)
            return processes

    def update(self, message=None):
        """
        将进度更新到下一个进度值
        @param message:
        @return:
        """
        self._update(self.processes.pop(0), message)

    def _update(self, current_process, message=None):
        if self.can_update(current_process) and current_process != self.last_process:
            if message is not None:
                self.update_callback(current_process, message)
            else:
                self.update_callback(current_process)
        self.last_process = current_process

    @contextmanager
    def hand_out(self, weight):
        """
        为子函数派生子进度条，子进度条根据权重分配父进度其中一个工作区块对应的进度区间
        @param weight:
        @return:
        """
        current_process = self.processes.pop(0)
        child = self.__class__(
            weight, self.update_callback,
            self.last_process, current_process, self._skip)
        yield child
        try:
            self._update(current_process)
        except TypeError:
            self._update(current_process, "")


class AsyncProcessor(Processor):
    """
    update方法变为异步的形式
    """

    async def update(self, message=None):
        """
        将进度更新到下一个进度值
        @param message:
        @return:
        """
        await self._update(self.processes.pop(0), message)

    async def _update(self, current_process, message=None):
        if self.can_update(current_process) and current_process != self.last_process:
            if message is not None:
                await self.update_callback(current_process, message)
            else:
                await self.update_callback(current_process)
        self.last_process = current_process

    @contextmanager
    async def hand_out(self, weight):
        """
        为子函数派生子进度条，子进度条根据权重分配父进度其中一个工作区块对应的进度区间
        @param weight:
        @return:
        """
        current_process = self.processes.pop(0)
        child = self.__class__(
            weight, self.update_callback,
            self.last_process, current_process, self._skip)
        yield child
        try:
            await self._update(current_process)
        except TypeError:
            await self._update(current_process, "")


"""
实例创建时，调用顺序
元类的__call__
类的__new__
类的__init__
子类的元类(除type以外)必须是父类元类的子类
使用了元类的类的子类也会使用该元类创建
"""
from threading import RLock


__all__ = ["Singleton"]


class Singleton(type):
    """
    单例类实现
    """
    lock = RLock()

    def __new__(mcs, *args, **kwargs):
        """
        元类msc通过__new__组建类对象，其中msc指Singleton
        :param args: 可以包含类构建所需要三元素，类名,父类,命名空间, 其中
        命名空间中__qualname__和函数的__qualname__均含有classname做为前缀，
        在这里，如果想替换类名，需要把以上全部替换才可以。
        :param kwargs: 可以自定义传递一些参数
        :return: 返回类对象,通过super(Singleton, mcs).__new__此时已经组装好了类
        """
        class_name, bases, dict = args
        dict["_instance"] = None
        cls = super(Singleton, mcs).__new__(
            mcs, class_name, bases, dict, **kwargs)
        return cls

    def __call__(cls, *args, **kwargs):
        with cls.lock:
            cls._instance = cls._instance or super(Singleton, cls).__call__(
                *args, **kwargs)
            return cls._instance


if __name__ == "__main__":

    class A(metaclass=Singleton):
        pass

    print(A() is A())


from collections.abc import MutableSequence, MutableMapping, MutableSet

__all__ = ["Frozen", "FrozenSettings"]


class Frozen(MutableSequence, MutableMapping):
    """
    替换json取数由[]变成.的形式
    """
    def __new__(cls, json):
        if isinstance(json, (MutableSequence, MutableSet)):
            instance = super().__new__(cls)
            object.__setattr__(instance, "_json",
                               json.__class__(cls(val) for val in json))
        elif isinstance(json, MutableMapping):
            instance = super().__new__(cls)
            object.__setattr__(instance, "_json", json)
        else:
            instance = json
        return instance

    def __getattr__(self, item):
        if hasattr(self._json, item):
            return getattr(self._json, item)
        else:
            prop_val = self._json.get(item)
            if prop_val is None:
                raise AttributeError(item)
            elif isinstance(prop_val, (list, dict)):
                return self.__class__(prop_val)
            else:
                return prop_val

    def __setattr__(self, key, value):
        raise NotImplementedError

    def __getitem__(self, item):
        return self._json[item]

    def __iter__(self):
        return iter(self._json)

    def __bool__(self):
        return bool(self._json)

    def __len__(self):
        return len(self._json)

    def __delitem__(self, item):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def insert(self, index, value):
        raise NotImplementedError

    def __str__(self):
        return str(self._json)

    __repr__ = __str__

    def normalize(self):
        """
        使用Frozen对象获取列表或字典时，为了支持级联`.`操作，
        字典和列表对象被转换成了Frozen对象，如果直接使用，有时可能会引发错误。
        使用此方法得到真实的列表或字典。
        :return:
        """
        val = self._json
        if isinstance(val, (MutableSequence, MutableSet)):
            return val.__class__(
                i.normalize() if hasattr(i, "normalize") else i for i in val)
        else:
            return val


class FrozenSettings(Frozen):
    pass


if __name__ == "__main__":
    json = {"aa": [1, 2, 3, {"b": 3, "c": [4, 5, {"d": 33}]}]}
    f = Frozen(json)
    print(f.aa[3].c[2].d)
    print(f)



# -*- coding:utf-8 -*-
import sys
import time

from socket import socket
from threading import Thread, local
from code import InteractiveInterpreter


__all__ = ["Consoler"]

_local = local()
_displayhook = sys.displayhook


class StringO(object):
    """
    替代标准输出
    """
    def __init__(self):
        self._buffer = []

    def isatty(self):
        return False

    def close(self):
        pass

    def flush(self):
        pass

    def seek(self, n, mode=0):
        pass

    def readline(self):
        if len(self._buffer) == 0:
            return ''
        ret = self._buffer[0]
        del self._buffer[0]
        return ret

    def reset(self):
        val = ''.join(self._buffer)
        del self._buffer[:]
        return val

    def _write(self, x):
        if isinstance(x, bytes):
            x = x.decode('utf-8', 'replace')
        self._buffer.append(str(x))

    def write(self, x):
        self._write(x)

    def writelines(self, x):
        self._write(''.join(x))


class ThreadedStream(object):

    """Thread-local wrapper for sys.stdout for the interactive console."""

    def push():
        if not isinstance(sys.stdout, ThreadedStream):
            sys.stdout = ThreadedStream()
        _local.stream = StringO()
    push = staticmethod(push)

    def fetch():
        try:
            stream = _local.stream
        except AttributeError:
            return ''
        return stream.reset()
    fetch = staticmethod(fetch)

    def displayhook(obj):
        try:
            stream = _local.stream
        except AttributeError:
            return _displayhook(obj)
        # stream._write bypasses escaping as debug_repr is
        # already generating HTML for us.
        if obj is not None:
            _local._current_ipy.locals['_'] = obj
            stream._write(obj)
    displayhook = staticmethod(displayhook)

    def __setattr__(self, name, value):
        raise AttributeError('read only attribute %s' % name)

    def __dir__(self):
        return dir(sys.__stdout__)

    def __getattribute__(self, name):
        if name == '__members__':
            return dir(sys.__stdout__)
        try:
            stream = _local.stream
        except AttributeError:
            stream = sys.__stdout__
        return getattr(stream, name)

    def __repr__(self):
        return repr(sys.__stdout__)


sys.displayhook = ThreadedStream.displayhook


class CustomInteractiveInterpreter(InteractiveInterpreter):

    def __init__(self, locals):
        super(CustomInteractiveInterpreter, self).__init__(locals)
        self.buffer = []
        self.stdout = sys.stdout
        self.more = False

    def runsource(self, source):
        source = source.rstrip() + '\n'
        try:
            ThreadedStream.push()
            source_to_eval = ''.join(self.buffer + [source])
            # print("exec source : %s"%source_to_eval)
            result = super(
                CustomInteractiveInterpreter, self).runsource(source_to_eval)
            # print("result: %s"%result)
            if (self.more or result) and source != "\n":
                self.buffer.append(source)
                self.more = True
            else:
                del self.buffer[:]
                self.more = False
        finally:
            output = ThreadedStream.fetch()
            sys.stdout = self.stdout
        prompt = self.more and '... ' or '>>> '
        return prompt + output

    def write(self, data):
        sys.stdout.write(data)


class Consoler(object):
    """
    通过交互客户端实时了解程序内部变化
    """
    parser = None
    alive = True

    def __init__(self, host, port, console, debug):
        if console:
            self._start_client(host, port)
        elif debug:
            self.namespace = locals()
            self._init_console(host, port)

    def _init_console(self, console_host, console_port):
        console_thread = Thread(
            target=self._console, args=(console_host, console_port))
        console_thread.setDaemon(True)
        console_thread.start()

    def _start_client(self, console_host, console_port):
        client = socket()
        client.connect((console_host, console_port))
        result = b""
        while self.alive:
            if not result.count(b"...") and result != b">>> ":
                print(">>> ", end="")
            try:
                cmd = input()
            except KeyboardInterrupt:
                break
            if not cmd:
                cmd = "\n"
            if cmd == "exit" or cmd == "exit()":
                break
            client.send(cmd.encode())
            result = client.recv(102400)
            if result:
                print(result.decode(),
                      end="" if result.count(b"...") or result == b">>> "
                      else "\n")
            else:
                break
        client.close()
        self.alive = False
        sys.exit(0)

    def _console(self, console_host, console_port):
        server = socket()
        server.setblocking(0)
        server.bind((console_host, console_port))
        server.listen(0)
        ci = CustomInteractiveInterpreter(self.namespace)
        _local._current_ipy = ci
        while self.alive:
            client = None
            try:
                client, addr = server.accept()
                while self.alive:
                    try:
                        cmd = client.recv(102400)
                        if not cmd:
                            break
                        result = ci.runsource(cmd.decode()).encode()
                        client.send(result)
                    except BlockingIOError as e:
                        time.sleep(0.1)
            except BlockingIOError as e:
                if not self.alive:
                    break
                time.sleep(1)
            finally:
                if client:
                    try:
                        client.close()
                    except:
                        pass
        server.close()
        self.alive = False

    def stop(self, *args):
        self.alive = False

    def enrich_parser_arguments(self):
        self.parser.add_argument(
            "--console", help="start a console. ", action="store_true")
        self.parser.add_argument(
            "--console-host", help="console host. ", default="")
        self.parser.add_argument(
            "--console-port", type=int, help="console port. ", default=7878)



# -*- coding:utf-8 -*-
import os
import re
import sys
import pdb
import time
import json
import types
import ctypes
import socket
import signal
import typing
import inspect
import logging
import warnings

from queue import Empty
from future.utils import raise_from
from functools import wraps, reduce

__version__ = '1.7.36'


def test_prepare(search_paths :typing.List[str]=None):
    """
    单元测试时，动态添加PYTHONPATH。

    :param search_paths: 要将哪个路径加入PYTHONPATH
    """
    if not search_paths:
        search_paths = [".."]

    for search_path in search_paths:
        sys.path.insert(0, os.path.abspath(search_path))
    del sys.modules["toolkit"]


def debugger():
    """
    pdb的简单封装，在debug=False的情况下，不会打断点
    """
    try:
        debug = bool(eval(os.environ.get("DEBUG", "0").lower().capitalize()))
    except Exception:
        debug = False

    if debug:
        d = pdb.Pdb()
        d.set_trace( sys._getframe().f_back)


def arg_to_iter(arg: typing.Any) -> typing.List[typing.Any]:
    """
    将非可迭代对象转换成可迭代对象

    :param arg: 任意对象
    :return: 列表对象
    """
    if arg is None:
        return []
    elif not isinstance(arg, (dict, str, bytes)) and hasattr(arg, '__iter__'):
        return arg
    else:
        return [arg]


class Compose(object):
    """
    函数聚合器
    """
    def __init__(self, *functions: typing.Tuple[types.FunctionType],
                 is_pipe: bool=False, aggregation: types.FunctionType=None):
        """
        连续调用多个函数, 如果is_pipe=True，依次连接所有函数的输出到输入。
        否则，返回所有函数结果集列表将传aggregation中进行聚合。

        :param functions: 需要连接的函数
        :param is_pipe: 是否建立管道
        :param aggregation: 聚合函数
        """
        self.functions = functions
        self.is_pipe = is_pipe
        if aggregation:
            self.aggregation = aggregation

    @staticmethod
    def aggregation(iterable: typing.Iterable) -> bool:
        """
        对所有运算结果进行逻辑与

        :param iterable: 所有运算结果集合
        :return: 运算结果True/False
        """
        return reduce(lambda x, y: x and y, iterable)

    def __call__(self, *args: typing.Tuple[typing.Any],
                 **kwargs: typing.Dict[typing.AnyStr, typing.AnyStr]):
        result_set = []
        for index, func in enumerate(self.functions):
            if not index or not self.is_pipe:
                result_set.append(func(*args, **kwargs))
            else:
                result_set[0] = func(result_set[0])
        return self.aggregation(
            result_set or [0, 0]) if len(result_set) != 1 else result_set[0]


def duplicate(iterable: typing.Iterable,
              keep: types.FunctionType=lambda x: x,
              key:types.FunctionType=lambda x: x,
              reverse: bool=False) -> list:
    """
    保序去重

    :param iterable: 需要去重的列表
    :param keep: 去重的同时要对element做的操作
    :param key: 使用哪一部分去重
    :param reverse: 是否反向去重
    :return:
    """
    result = list()
    duplicator = list()
    if reverse:
        iterable = reversed(iterable)

    for i in iterable:
        keep_field = keep(i)
        key_words = key(i)
        if key_words not in duplicator:
            result.append(keep_field)
            duplicator.append(key_words)
    return list(reversed(result)) if reverse else result


def strip(value, chars: typing.AnyStr=None):
    """
    安全strip，如果不是字符串类型，直接返回。

    :param value: 要strip的值
    :param chars: 需要去掉的字符
    :return: 结果
    """
    if isinstance(value, (bytes, str)):
        return value.strip(chars)
    return value


def decode(value, encoding: str="utf-8"):
    """
    decode字段

    :param value: 要decode的值
    :param encoding: 编码
    :return: 结果
    """
    return value.decode(encoding)


def encode(value, encoding: str="utf-8"):
    """
    encode字段

    :param value: 要encode的值
    :param encoding: 编码
    :return: 结果
    """
    return value.encode(encoding)


def rid(value: typing.AnyStr,
        old: typing.Union[typing.AnyStr, typing.Pattern[typing.AnyStr]],
        new: typing.AnyStr) -> typing.AnyStr:
    """
    去掉匹配成功的字段

    :param value: 要被处理的字符串
    :param old: 被替换的内容，可以为正则表达示或字符串
    :param new: 替换的字符串
    :return: 结果
    """
    if hasattr(old, "sub"):
        return old.sub(new, value)
    else:
        return value.replace(old, new)


def wrap_key(json_str: str,
             key_pattern: typing.Pattern[str]=re.compile(r"([a-zA-Z_]\w*)[\s]*:")
             ) -> str:
    """
    将javascript 对象字串串形式的key转换成被双字符包裹的格式如{a: 1} => {"a": 1}

    :param json_str: 要处理的json字符串
    :param key_pattern: 用来处理的正则表达式
    :return: 结果
    """
    return key_pattern.sub('"\g<1>":', json_str)


def safely_json_loads(json_str: typing.AnyStr,
                      defaulttype: typing.Any=dict,
                      escape: bool=True):
    """
    返回安全的json类型

    :param json_str: 要被loads的字符串
    :param defaulttype: 若load失败希望得到的对象类型
    :param escape: 是否将单引号变成双引号
    :return: 结果
    """
    if not json_str:
        return defaulttype()
    elif escape:
        data = replace_quote(json_str)
        return json.loads(data)
    else:
        return json.loads(json_str)


def chain_all(iter: typing.Iterable) -> typing.Union[typing.Collection]:
    """
    连接多个序列或字典

    :param iter: 字典或列表的列表
    :return: 结果
    """
    iter = list(iter)

    if not iter:
        return []

    if isinstance(iter[0], dict):
        result = {}
        for i in iter:
            result.update(i)
    else:
        result = reduce(lambda x, y: list(x)+list(y), iter)
    return result


def replace_quote(json_str: str) -> str:
    """
    将要被json.loads的字符串的单引号转换成双引号，
    如果该单引号是元素主体，而不是用来修饰字符串的。则不对其进行操作。

    :param json_str: json字符串
    :return: 结果
    """
    if not isinstance(json_str, str):
        return json_str

    double_quote = []
    new_lst = []

    for index, val in enumerate(json_str):
        if val == '"' and not len(_get_last_backslash(json_str[:index])) % 2:
            if double_quote:
                double_quote.pop(0)
            else:
                double_quote.append(val)

        if val == "'" and not len(_get_last_backslash(json_str[:index])) % 2:
            if not double_quote:
                val = '"'
        new_lst.append(val)
    return "".join(new_lst)


def _get_last_backslash(strings, regex=re.compile(r"\\*$")):
    mth = regex.search(strings)
    if mth:
        return mth.group()
    return ""


def format_html_string(html: str) -> str:
    """
    格式化html, 去掉多余的字符，类，script等。这个函数性能可能会较差。

    :param html: 传入的html字符串
    :return: 处理完的html
    """
    trims = [(r'\n',''),
             (r'\t', ''),
             (r'\r', ''),
             (r'  ', ''),
             (r'\u2018', "'"),
             (r'\u2019', "'"),
             (r'\ufeff', ''),
             (r'\u2022', ":"),
             (r"<([a-z][a-z0-9]*)\ [^>]*>", '<\g<1>>'),
             (r'<\s*script[^>]*>[^<]*<\s*/\s*script\s*>', ''),
             (r"</?a.*?>", '')]
    return reduce(lambda string, replacement:
                  re.sub(replacement[0], replacement[1], string), trims, html)


def urldecode(query: str) -> dict:
    """
    与urlencode相反，不过没有unquote

    :param query: url查询字符串
    :return: 查询字符串的字典
    """
    if not query.strip():
        return dict()

    return dict(x.split("=", 1) for x in query.strip().split("&"))


def re_search(regex: typing.Union[typing.Pattern[str], str],
              text: typing.AnyStr,
              dotall: bool=True,
              default: str="") -> str:
    """
    抽取正则规则的第一组元素

    :param regex: 正则对象或字符串
    :param text: 被查找的字符串
    :param dotall: 正则.是否匹配所有字符
    :param default: 找不到时的默认值
    :return: 抽取正则规则的第一组
    """
    if isinstance(text, bytes):
        text = text.decode("utf-8")
    if not isinstance(regex, list):
        regex = [regex]
    for rex in regex:
        rex = (re.compile(rex, re.DOTALL)
               if dotall else re.compile(rex)) if isinstance(rex, str) else rex
        match_obj = rex.search(text)
        if match_obj is not None:
            t = match_obj.group(1).replace('\n', '')
            return t
    return default


class P22P3Encoder(json.JSONEncoder):
    """
    python2转换python3时使用的json encoder
    """

    def default(self, obj):
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        if isinstance(obj, (types.GeneratorType, map, filter)):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def retry_wrapper(
        retry_times: int,
        exception: type=Exception,
        error_handler: types.FunctionType=None,
        interval: float=0.1):
    """
    函数出错时用来重试的装饰器

    :param retry_times: 重试次数
    :param exception: 需要重试的异常
    :param error_handler: 出错时的回调函数
    :param interval: 重试间隔时间
    """
    def out_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            count = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except exception as e:
                    count += 1
                    if error_handler:
                        result = error_handler(
                            func.__name__, count, e, *args, **kwargs)
                        if result:
                            count -= 1
                    if count >= retry_times:
                        raise
                    time.sleep(interval)
        return wrapper
    return out_wrapper


def timeout(timeout_time: int, default):
    """
    装饰一个方法，用来限制其执行时间，超时后返回default，只能在主线程使用。

    :param timeout_time: 超时时间
    :param default: 超时后的返回
    """
    class DecoratorTimeout(Exception):
        pass

    def timeout_function(f):
        def f2(*args):
            def timeout_handler(signum, frame):
                raise DecoratorTimeout()

            old_handler = signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(timeout_time)
            try:
                retval = f(*args)
            except DecoratorTimeout:
                return default
            finally:
                signal.signal(signal.SIGALRM, old_handler)
            signal.alarm(0)
            return retval
        return f2
    return timeout_function


def custom_re(regex: typing.Pattern[typing.AnyStr], text: typing.AnyStr) -> list:
    """
    模仿selector.re

    :param regex: 正则表达式对象
    :param text: 被查询的字符串
    :return:
    """
    return re.findall(regex, text)


def replace_dot(data: dict) -> dict:
    """
    mongodb不支持key中带有"."，该函数用来将"."转换成"_"

    :param data: 要处理的字典
    :return: 返回的字典
    """
    return dict((k.replace(".", "_"), v) for k, v in data.items())


def groupby(it: typing.Iterable, key: types.FunctionType) -> dict:
    """
    自实现groupby，itertool的groupby不能合并不连续但是相同的组, 且返回值是iter

    :param it: 一个列表套字典的结构
    :param key: 从字典从取值的函数，作为group_by后的key
    :return: 字典对象
    """
    groups = dict()
    for item in it:
        groups.setdefault(key(item), []).append(item)
    return groups


def parse_cookie(string: str,
                 regex: typing.Pattern[str]=re.compile(r'([^=]+)=([^\;]+);?\s?')
                 ) -> dict:
    """
    解析cookie

    :param string: cookie字符串
    :param regex: 解析cookie的正则表达式
    :return: cookie字典
    """
    return dict((k, v) for k, v in regex.findall(string))


def async_produce_wrapper(producer, logger, batch_size=10):
    """
    pykafka实现异步生产时，使用的装饰器
    `
    self.producer.produce =
    async_produce_wrapper(self.producer, self.logger)(self.producer.produce)
    `
    :param producer:
    :param logger:
    :param batch_size:
    :return:
    """
    count = 0

    def wrapper(func):

        def inner(*args, **kwargs):
            result = func(*args, **kwargs)
            nonlocal count
            count += 1
            if count % batch_size == 0:  # adjust this or bring lots of RAM ;)
                while True:
                    try:
                        msg, exc = producer.get_delivery_report(block=False)
                        if exc is not None:
                            logger.error('Failed to deliver msg {}: {}'.format(
                                msg.partition_key, repr(exc)))
                        else:
                            logger.info('Successfully delivered msg {}'.format(
                                msg.partition_key))
                    except Empty:
                        break
            return result
        return inner
    return wrapper


def load_function(function_str: str) -> types.FunctionType:
    """
    返回字符串表示的函数对象

    :param function_str: 函数描述字符串，如：module1.module2.function
    :return: function: 找到的函数
    """
    warnings.warn("load_function() is a deprecated alias since 1.7.19, "
                  "use load() instead.",
                  DeprecationWarning, 2)
    mod_str, _sep, function_str = function_str.rpartition('.')
    if mod_str:
        return getattr(load_module(mod_str), function_str)
    else:
        # 可能整体是一个模块
        return load_module(function_str)
    

load_class = load_function


def load_module(module_str: str) -> types.ModuleType:
    """
    返回字符串表示的模块

    :param module_str: 模块描述字符串，如: os.path
    :return: 模块，如: os.path
    """
    warnings.warn("load_function() is a deprecated alias since 1.7.19, "
                  "use load() instead.",
                  DeprecationWarning, 2)
    return __import__(module_str, fromlist=module_str.split(".")[-1])


def load(prop_str: str):
    """
    返回字符串表示的模块、函数、类、若类的属性等

    :param prop_str: 函数、类、类属性、模块描述字符串，如: module1.class.function....
    :return: 函数、类、模块等。
    """
    attr_list = []
    # 每次循环将prop_str当模块路径查找，成功则返回，
    # 失败则将模块路径回退一级，将回退的部分转换成属性
    # 至到加载模块成功后依次从模块中提取属性。
    ex = None
    while prop_str:
        try:
            obj = __import__(prop_str, fromlist=prop_str.split(".")[-1])
            for attr in attr_list:
                obj = getattr(obj, attr)
            return obj
        except (AttributeError, ImportError) as e:
            prop_str, _sep, attr_str = prop_str.rpartition('.')
            attr_list.insert(0, attr_str)
            try:
                raise_from(ImportError(e), e)
            except Exception as e:
                ex = e
    else:
        raise ex


def free_port() -> int:
    """
    找到一个可用端口

    :return 端口号
    """
    free_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    free_socket.bind(('0.0.0.0', 0))
    free_socket.listen(5)
    port = free_socket.getsockname()[1]
    free_socket.close()
    return port


def thread_safe(lock):
    """
    对指定函数进行线程安全包装，需要提供锁

    :param lock: 锁
    """
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with lock:
                return func(*args, **kwargs)
        return wrapper
    return decorate


def thread_safe_for_method(func):
    """
    对类中的方法进行线程安全包装
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        with args[0].lock:
            return func(*args, **kwargs)
    return wrapper


def call_later(callback: str,
               call_args: typing.Tuple=tuple(),
               immediately: bool=True,
               interval: typing.Union[float, int]=1):
    """
    应用场景：
    被装饰的方法需要大量调用，随后需要调用保存方法，
    但是因为被装饰的方法访问量很高，而保存方法开销很大
    所以设计在装饰方法持续调用一定间隔后，再调用保存方法。
    规定间隔内，无论调用多少次被装饰方法，保存方法只会
    调用一次，除非immediately=True

    :param callback: 随后需要调用的方法名
    :param call_args: 随后需要调用的方法所需要的参数
    :param immediately: 是否立即调用
    :param interval: 调用间隔
    """
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self = args[0]
            try:
                return func(*args, **kwargs)
            finally:
                if immediately:
                    getattr(self, callback)(*call_args)
                else:
                    now = time.time()
                    if now - self.__dict__.get("last_call_time", 0) > interval:
                        getattr(self, callback)(*call_args)
                        self.__dict__["last_call_time"] = now
        return wrapper
    return decorate


def get_ip() -> str:
    """
    获取局域网ip

    :return: 局域网ip
    """
    try:
        import psutil
    except ImportError:
        warnings.warn(
            "get_ip function depends on psutil, try: pip install psutil. ")
        raise
    netcard_info = []
    info = psutil.net_if_addrs()
    for k,v in info.items():
        for item in v:
            if item[0] == 2 and not item[1]=='127.0.0.1':
                netcard_info.append((k,item[1]))

    if netcard_info:
        return netcard_info[0][1]


def find_caller_name(is_func=False, steps=1):
    """
    用来发现调用这个方法的调用者的名字，类名或模块名
    :param is_func:
    :param steps:
    :return:
    """
    frame = logging.currentframe()
    for i in range(steps + 1):
        co = frame.f_code
        frame = frame.f_back

    if is_func:
        return co.co_name
    else:
        filename = os.path.basename(co.co_filename).split(".")[0]
        if filename == "__init__":
            filename = os.path.basename(os.path.dirname(co.co_filename))
        return filename

#
# class LazyDict(object):
#     """
#     懒加载dict, 提供一个转换函数，只有在获取dict中的值时，才对指定值进行turn函数的调用
#     """
#     def __init__(self, d, turn):
#         self.dict = d
#         self.turn = turn
#
#     def get(self, item):
#         return self.dict.setdefault(item, self.turn(item, self.dict))
#
#     def __getitem__(self, item):
#         return self.get(item)
#
#     def to_dict(self):
#         return self.dict


def cache_property(func):
    """
    缓存属性，只计算一次
    """
    return property(_property_cache(func))


def _property_cache(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if func.__name__ not in args[0].__dict__:
            args[0].__dict__[func.__name__] = func(*args, **kwargs)
        return args[0].__dict__[func.__name__]

    return wrapper


class cache_prop(object):
    """
    描述符版缓存属性，无法被覆盖
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        if self.func.__name__ in instance.__dict__:
            return instance.__dict__[self.func.__name__]
        else:
            return instance.__dict__.setdefault(
                self.func.__name__, self.func(instance))

    def __set__(self, instance, value):
        raise AttributeError("{} is readonly.".format(self.func.__name__))


def cache_for(timeout: typing.Union[float, int]=10):
    """
    缓存属性，指定缓存失效时间

    :param timeout: 缓存失效时间 second
    """
    def outer(func):
        @property
        @wraps(func)
        def inner(*args, **kwargs):
            prop_name = func.__name__
            prop_start_name = "%s_cache_start_time" % prop_name
            if time.time() - args[0].__dict__.get(prop_start_name, 0) > timeout:
                args[0].__dict__[prop_name] = func(*args, **kwargs)
                args[0].__dict__["%s_cache_start_time" % prop_name] = time.time()
            return args[0].__dict__[prop_name]
        return inner
    return outer


def free_namedtuple(func):
    """
    根据一个类及其参数创建一个类namedtuple class，
    但不同之处在于创建实例成功后可以自由赋值，初时化时指定的值决定其Hash和eq结果

    :param func:
    :return:
    """
    cls_tmpl = """
class {}(object):
    def __init__(self, {}):
        {}
    def __hash__(self):
        return {}

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return {}
"""

    args = inspect.getargspec(func).args
    if args and args[0] in ["self", "cls"]:
        args.pop(0)

    class_name = func.__name__.capitalize()
    init_arg = ", ".join(args + "=None" for args in args) if args else ""
    init_body = "".join(
        "self.{%s} = {%s}\n        " % (
            index, index) for index in range(len(args))).format(*args) if args else "pass"
    hash_body = " + ".join("hash(self.{})".format(arg) for arg in args) if args else "0"
    eq_body = " and ".join("self.{0} == other.{0}".format(arg) for arg in args) if args else "True"
    namespace = dict(__name__='entries_%s' % class_name)
    class_str = cls_tmpl.format(class_name, init_arg, init_body, hash_body, eq_body)
    exec(class_str, namespace)
    return namespace[class_name]


def cache_method(timeout: int=10):
    """
    缓存一个方法的调用结果，持续一定时长，
    根据不同的调用参数来缓存不同的结果，调用参数必须是可hash的，
    该方法调用失败时，希望抛出异常，此时缓存会失败。

    :param timeout: 缓存时长 :s
    """
    def cache(func):
        entry_class = free_namedtuple(func)
        data = dict()

        @wraps(func)
        def inner(*args, **kwargs):
            for k in data.copy().keys():
                if time.time() - timeout > data[k].ts:
                    del data[k]

            entry = entry_class(*args[1:], **kwargs)
            entry.ts = time.time()
            stored = data.get(entry, entry)
            if entry not in data or time.time() - timeout > stored.ts:
                entry.result = func(*args, **kwargs)
                data[entry] = entry
                return entry.result
            else:
                return stored.result
        return inner
    return cache


def cache_method_for_update(func: types.FunctionType):
    """
    缓存方法调用结果直到实例的updated返回True，与cache_method类似。

    @param func: 要缓存的方法
    """
    entry_class = free_namedtuple(func)
    data = dict()

    @wraps(func)
    def inner(*args, **kwargs):
        self = args[0]
        # 没有更新方法(直接使用子类，而不通过cache)，不进行缓存操作
        if not self.updated:
            return func(*args, **kwargs)
        # 如果已更新，则清除所有缓存
        if self.updated():
            data.clear()
        entry = entry_class(*args[1:], **kwargs)
        stored = data.get(entry, entry)

        if entry not in data:
            entry.result = func(*args, **kwargs)
            if entry.result:
                data[entry] = entry
            return entry.result
        else:
            return stored.result

    return inner

#
# class SafeValue(object):
#     """
#     获取某个对象的某个属性值，如果该属性不存在，则返回该对象
#     """
#     def __init__(self, node):
#         self.node = node
#
#     def __getattr__(self, item):
#         return getattr(self.node, item, self.node)


def overflow(val: int, btc: int) -> int:
    """
    对于超出字节长度的数字转换成负数

    :param val: 值
    :param btc: 字节长度
    """
    max_val = 2 ** (8*btc-1) - 1
    if not -max_val - 1 <= val <= max_val:
        val = (val + (max_val + 1)) % (2 * (max_val + 1)) - max_val - 1
    return val


def int_overflow(val: int) -> int:
    """
    类似js对于超过int范围的数转换成负数
    :param val: 值
    :return: 结果
    """
    return overflow(val, 4)


def unsigned_right_shift(n: int, i: int) -> int:
    """
    无符号右移python实现

    :param n: 值
    :param i: 移的位数
    :return: 结果
    """
    # 数字小于0，则转为32位无符号uint
    if n < 0:
        n = ctypes.c_uint32(n).value
    # 正常位移位数是为正数，但是为了兼容js之类的，负数就右移变成左移好了
    if i < 0:
        return -int_overflow(n << abs(i))
    return int_overflow(n >> i)


def shift_left_for_js(num: int, count: int) -> int:
    """
    位运算左移js版

    :param num: 值
    :param count: 移的位数
    :return: 结果
    """
    return int_overflow(num << count)


def shift_right_for_js(num: int, count: int) -> int:
    """
    位运算右移js版

    :param num: 值
    :param count: 移的位数
    :return: 结果
    """
    return int_overflow(num >> count)


async def readexactly(stream, n: int) -> bytes:
    """
    asyncio stream read

    :param stream: 流
    :param n: 读取字节的长度
    :return:
    """
    if hasattr(stream, "_exception") and stream._exception is not None:
        raise stream._exception

    blocks = []
    while n > 0:
        block = await stream.read(n)
        if not block:
            break
        blocks.append(block)
        n -= len(block)

    return b''.join(blocks)


class NotImplementedProp(object):
    """
    用来对子类需要实现的类属性进行占位
    """

    def __get__(self, instance, owner):
        return NotImplemented


def find_ancestor(cls, until_not_have=None):
    """
    查找祖先，直到祖先为object或者祖先没有until_not_have这个属性时，返回当前类。

    :param cls:
    :return:
    """
    if until_not_have and not hasattr(cls.__base__, until_not_have) \
            or cls.__base__ == object:
        return cls
    return find_ancestor(cls.__base__)


class classproperty(object):
    """
    property只能用于实例方法到实例属性的转换，使用classproperty来支持类方法到类属性的转换
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        return self.func(owner)


def cache_classproperty_for(all=False):
    """
    缓存类属性，只计算一次，如果all=True，子类和父类共用

    :param all:
    :return:
    """
    def _cache_classproperty(func: types.FunctionType):

        base = None

        @classproperty
        @wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal base
            if base is None:
                if all:
                    base = find_ancestor(args[0], until_not_have=func.__name__)
                else:
                    base = args[0]

            prop_name = "_" + func.__name__
            if prop_name not in base.__dict__:
                setattr(base, prop_name, func(*args, **kwargs))
            return base.__dict__[prop_name]

        return wrapper
    return _cache_classproperty


global_cache_classproperty = cache_classproperty_for(all=True)


cache_classproperty = cache_classproperty_for(all=False)


def _assert(boolean, data):
    """
    增强版assert，如果data是异常对象则直接抛出。

    :param boolean: True/False
    :param data: 异常或者描述信息
    """
    if not boolean:
        if isinstance(data, Exception):
            raise data
        raise AssertionError(data)



import uuid
import time


class DistributedLock(object):
    """
    Redis分布式锁
    """
    SCRIPT = "if redis.call('get', KEYS[1]) == ARGV[1] then " \
             "return redis.call('del', KEYS[1]) else return 0 end"

    def __init__(self, redis_conn, lock_key, expire, roll_interval=0.1):
        self.lock_key = lock_key
        self.redis_conn = redis_conn
        self.req_id = uuid.uuid4()
        self.expire = expire
        self.roll_interval = roll_interval

    def __enter__(self):
        self.lock()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unlock()

    async def __aenter__(self):
        return await self.alock()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.aunlock()

    def _lock(self):
        return self.redis_conn.set(
            self.lock_key, self.req_id, ex=self.expire, nx=True)

    def lock(self):
        while not self._lock():
            time.sleep(self.roll_interval)

    def unlock(self):
        script = self.redis_conn.register_script(self.SCRIPT)
        return script(keys=[self.lock_key], args=[self.req_id])

    async def _alock(self):
        return await self.redis_conn.set(
            self.lock_key, self.req_id, ex=self.expire, nx=True)

    async def alock(self):
        while not await self._alock():
            time.sleep(self.roll_interval)

    async def aunlock(self):
        script = self.redis_conn.register_script(self.SCRIPT)
        return await script.execute(keys=[self.lock_key], args=[self.req_id])



# -*- coding:utf-8 -*-
import os
import re
import string

from contextlib import contextmanager
from setuptools import setup, find_packages


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    mth = re.search("__version__\s?=\s?['\"]([^'\"]+)['\"]", init_py)
    if mth:
        return mth.group(1)
    else:
        raise RuntimeError("Cannot find version!")


def install_requires():
    """
    Return requires in requirements.txt
    :return:
    """
    try:
        with open("requirements.txt") as f:
            return [line.strip() for line in f.readlines() if line.strip()]
    except OSError:
        return []


VERSION = get_version("toolkit")

AUTHOR = "cn"

AUTHOR_EMAIL = "cnaafhvk@foxmail.com"

URL = "https://www.github.com/ShichaoMa/toolkit"

NAME = "toolkity"

DESCRIPTION = "simple function tools. "

try:
    LONG_DESCRIPTION = open("README.rst").read()
except UnicodeDecodeError:
    LONG_DESCRIPTION = open("README.rst", encoding="utf-8").read()

KEYWORDS = "tools function"

LICENSE = "MIT"


@contextmanager
def cfg_manage(cfg_tpl_filename):
    if os.path.exists(cfg_tpl_filename):
        cfg_file_tpl = open(cfg_tpl_filename)
        buffer = cfg_file_tpl.read()
        try:
            with open(cfg_tpl_filename.rstrip(".tpl"), "w") as cfg_file:
                cfg_file.write(string.Template(buffer).substitute(
                    pwd=os.path.abspath(os.path.dirname(__file__))))
            yield
        finally:
            cfg_file_tpl.close()
    else:
        yield


with cfg_manage(__file__.replace(".py", ".cfg.tpl")):
    setup(
        name=NAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python',
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
        ],
        entry_points={
            "console_scripts": [
                "ver-inc = toolkit.package_control:change_version",
                "mk-html = toolkit.markdown_helper:main",
            ]
        },
        keywords=KEYWORDS,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        url=URL,
        license=LICENSE,
        packages=find_packages(exclude=("tests*",)),
        install_requires=install_requires(),
        include_package_data=True,
        zip_safe=True,
        setup_requires=["pytest-runner"],
        tests_require=["pytest-apistellar", "pytest-asyncio", "pytest-cov"]
    )


"""
settings模块使用方法
In [3]: from toolkit.settings import SettingsLoader

In [4]: sw = SettingsLoader()

In [5]: settings = sw.load({"a": 1}， "settings")

In [6]: settings
Out[6]: {'AUTHOR': '夏洛之枫', 'DB': 'sqlite', 'a': 1}

In [8]: settings.a
Out[8]: 1

In [9]: settings["a"]
Out[9]: 1

In [10]: settings.get("a")
Out[10]: 1

In [13]: import os

In [15]: os.environ["b"] = "3"

In [16]: settings.get_int("b", 11)
Out[16]: 3

In [18]: settings2 = sw.load({"a": 1, "c": {"d": [3,4,5,6], "e": {"a", 4, 5,6}}})

In [19]: settings2 is settings
Out[19]: False

In [21]: settings2.a
Out[21]: 1

In [22]: settings2.b
Out[22]: '3'

In [23]: settings2.c
Out[23]: {'d': [3, 4, 5, 6], 'e': {4, 5, 6, 'a'}}

In [24]: type(settings2.c)
Out[24]: toolkit.frozen.Frozen

In [26]: settings2.c.e
Out[26]: {4, 5, 6, 'a'}

In [27]: type(settings2.c.e)
Out[27]: set

In [28]: type(settings2.c.d)
Out[28]: toolkit.frozen.Frozen
"""
import os
import copy
import types
import logging
import importlib

from numbers import Number
from collections import UserDict
from collections.abc import Collection, MutableSequence, MutableSet, MutableMapping

from . import duplicate
from .frozen import FrozenSettings

__all__ = ["Settings", "SettingsLoader", "FrozenSettings"]

logger = logging.getLogger("toolkit.settings")


class Settings(UserDict):
    """
    settings，首先从环境变量从获取
    """
    def __getitem__(self, item):
        if item in os.environ:
            return os.environ[item]
        return super(Settings, self).__getitem__(item)

    def get(self, k, d=None):
        if k in os.environ:
            return os.environ[k]
        return super(Settings, self).get(k, d)

    def get_int(self, k, d=None):
        try:
            return int(self.get(k, d))
        except TypeError:
            return d

    def get_bool(self, k, d=False):
        try:
            try:
                return eval(self.get(k, str(d)))
            except:
                pass
            return bool(self.get(k, d))
        except TypeError:
            return d

    def get_float(self, k, d=None):
        try:
            return float(self.get(k, d))
        except TypeError:
            return d


class SettingsLoader(object):
    """
    配置文件加载装饰器用来加载和覆盖配置信息
    """
    ignore = [
        '__builtins__',
        '__file__',
        '__package__',
        '__doc__',
        '__name__',
        '__cached__',
    ]
    allow_types = [Collection, Number]

    def __init__(self, settings_type=Settings):
        self.settings = settings_type()

    @classmethod
    def register_types(cls, types):
        cls.allow_types.extend(types)

    def load(self, local='localsettings', default='settings'):
        """
        加载配置字典
        :param local: 本地配置模块
        :param default: 默认配置模块
        :return: 配置信息字典
        """
        for settings in [default, local]:
            if isinstance(settings, MutableMapping):
                self.merge(self.settings, settings)
            else:
                self._load(settings)

        return FrozenSettings(self.settings)

    def load_from_string(self, settings_string='', module_name='customsettings'):
        """
        从配置语句中获取配置信息
        :param settings_string: 配置信息语句
        :param module_name: 配置信息的环境变量
        :return:
        """
        try:
            mod = types.ModuleType(module_name)
            exec(settings_string, mod.__dict__)
        except TypeError:
            logger.warning("Could not import settings from string.")
            mod = None
        try:
            self.merge(self.settings, self._convert_to_dict(mod))
        except ImportError:
            logger.warning("Settings from string unable to be loaded.")

    def _load(self, setting_module):
        try:
            if isinstance(setting_module, str):
                if setting_module[-3:] == '.py':
                    setting_module = setting_module[:-3].replace(
                        os.sep, ".").strip(".")
                settings = importlib.import_module(setting_module)
            else:
                settings = setting_module
            self.merge(self.settings, self._convert_to_dict(settings))
        except ImportError:
            logger.warning("Cannot found %s." % setting_module)

    def merge(self, settings, new_settings):
        """
        合并settings
        :param settings:
        :param new_settings:
        :return:
        """
        for key in new_settings:
            value = settings.get(key)
            new_value = new_settings[key]
            if type(value) == type(new_value):
                if isinstance(value, MutableMapping):
                    self.merge(value, new_value)
                elif isinstance(value, MutableSet):
                    for v in new_value:
                        value.add(v)
                elif isinstance(value, MutableSequence):
                    value.extend(new_value)
                    settings[key] = duplicate(value, reverse=True)
                else:
                    settings[key] = new_value
            else:
                settings[key] = new_value

    def _convert_to_dict(self, settings):
        """
        将配置文件转化为字典
        :param setting:
        :return:
        """
        allow_types = tuple(self.allow_types)
        return dict(
            (k, copy.deepcopy(getattr(settings, k))) for k in dir(settings)
            if k not in self.ignore and isinstance(
                getattr(settings, k), allow_types))



import sys
import asyncio
import inspect

from functools import wraps
from threading import Thread
from contextlib import _GeneratorContextManager

from . import _property_cache


def contextmanager(func):
    """
    异步下上文管理器，其支持装饰同步生成器或异步生成器，并支持使用async with 或with。
    :param func:
    :return:
    """
    @wraps(func)
    def helper(*args, **kwds):
        return _AsyncGeneratorContextManager(func, args, kwds)

    return helper


def await_async_method(method, args: tuple=None, kwargs: dict=None):
    """
    如果event_loop已经开始，那么在异步方法中执行同步方法时又遇到了异步方法，
    使用run_until_complete会报错: `This event loop is already running`
    所以启用一个子线程来完成这件事。
    :param method:
    :param args:
    :param kwargs:
    :return:
    """
    container = dict()
    th = Thread(target=_async_run_in_child_thread_loop,
                args=(container, method, args, kwargs))
    th.start()
    th.join()
    if "rs" in container:
        return container["rs"]
    raise container["err"]


def sync_run_async(func):
    """
    同步执行异步
    :param func:
    :return:
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return await_async_method(func, args=args, kwargs=kwargs)

    return wrapper


def async_property(func):
    """
    异步property，支持将异步方法转换成同步属性，但不支持后续的property.setter等方法。
    :return:
    """
    return property(sync_run_async(func))


def async_cached_property(func):
    """
    有缓存效果异步属性
    :param func:
    :return:
    """
    return property(_property_cache(sync_run_async(func)))


def _async_run_in_child_thread_loop(container, method, args, kwargs):
    """
    使用一子线程事件循环还执行异步方法。
    :param container:
    :param method:
    :param args:
    :param kwargs:
    :return:
    """
    loop = asyncio.new_event_loop()
    try:
        args = args or tuple()
        kwargs = kwargs or dict()
        container["rs"] = loop.run_until_complete(method(*args, **kwargs))
    except Exception as e:
        container["err"] = e


class _AsyncGeneratorContextManager(_GeneratorContextManager):
    """
    异步上下文管理器
    """
    def __enter__(self):
        try:
            if inspect.isasyncgen(self.gen):
               return await_async_method(self.gen.asend, args=(None, ))
            else:
                return self.gen.send(None)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __exit__(self, type, value, traceback):
        if type is None:
            try:
                if inspect.isgenerator(self.gen):
                    self.gen.send(None)
                elif inspect.isasyncgen(self.gen):
                    await_async_method(self.gen.asend, args=(None, ))
                else:
                    return False
            except (StopIteration, StopAsyncIteration):
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                # Need to force instantiation so we can reliably
                # tell if we get the same exception back
                value = type()
            try:
                if inspect.isgenerator(self.gen):
                    self.gen.throw(type, value, traceback)
                elif inspect.isasyncgen(self.gen):
                    await_async_method(
                        self.gen.athrow, args=(type, value, traceback))

                else:
                    raise value
            except (StopIteration, StopAsyncIteration) as exc:
                # Suppress StopIteration *unless* it's the same exception that
                # was passed to throw().  This prevents a StopIteration
                # raised inside the "with" statement from being suppressed.
                return exc is not value
            except RuntimeError as exc:
                # Don't re-raise the passed in exception. (issue27122)
                if exc is value:
                    return False
                # Likewise, avoid suppressing if a StopIteration exception
                # was passed to throw() and later wrapped into a RuntimeError
                # (see PEP 479).
                if type is StopIteration and exc.__cause__ is value:
                    return False
                raise
            except:
                # only re-raise if it's *not* the exception that was
                # passed to throw(), because __exit__() must not raise
                # an exception unless __exit__() itself failed.  But throw()
                # has to raise the exception to signal propagation, so this
                # fixes the impedance mismatch between the throw() protocol
                # and the __exit__() protocol.
                #
                if sys.exc_info()[1] is value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")

    async def __aenter__(self):
        try:
            if inspect.isgenerator(self.gen):
                return self.gen.send(None)
            else:
                return await self.gen.asend(None)
        except (StopAsyncIteration, StopIteration):
            raise RuntimeError("generator didn't yield") from None

    async def __aexit__(self, type, value, traceback):
        if type is None:
            try:
                if inspect.isgenerator(self.gen):
                    self.gen.send(None)
                else:
                    await self.gen.asend(None)
            except (StopAsyncIteration, StopIteration):
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                value = type()
            try:
                if inspect.isgenerator(self.gen):
                    await self.gen.throw(type, value, traceback)
                else:
                    await self.gen.athrow(type, value, traceback)
            except (StopAsyncIteration, StopIteration) as exc:
                return exc is not value
            except RuntimeError as exc:
                if exc is value:
                    return False
                if issubclass(type, (StopAsyncIteration, StopIteration))\
                        and exc.__cause__ is value:
                    return False
                raise
            except:
                if sys.exc_info()[1] is value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")



import os
import sys
import logging
import datetime

from functools import wraps
from logging import handlers
from threading import current_thread
from pythonjsonlogger.jsonlogger import JsonFormatter

from . import find_caller_name
from .singleton import Singleton

__all__ = ["UDPLogstashHandler", "Logger"]


class UDPLogstashHandler(handlers.DatagramHandler):
    """Python logging handler for Logstash. Sends events over UDP."""
    def makePickle(self, record):
        return self.formatter.format(record).encode()


class Logger(object, metaclass=Singleton):
    """
    logger的实现
    """
    format_string = '%(asctime)s [%(name)s:%(threadname)s]%(level)s: %(message)s'

    def __init__(self, settings, name=None):
        self.name = name or find_caller_name(steps=2)
        self.json = settings.get_bool('LOG_JSON', True)
        self.level = settings.get('LOG_LEVEL', 'DEBUG')
        self.stdout = settings.get_bool('LOG_STDOUT', True)
        self.dir = settings.get('LOG_DIR', 'logs')
        self.bytes = settings.get('LOG_MAX_BYTES', 10*1024*1024)
        self.backups = settings.get_int('LOG_BACKUPS', 5)
        self.udp_host = settings.get("LOG_UDP_HOST", "127.0.0.1")
        self.udp_port = settings.get_int("LOG_UDP_PORT", 5230)
        self.log_file = settings.get_bool('LOG_FILE', False)
        self.logger = logging.getLogger(self.name)
        self.logger.propagate = False
        self.set_up()

    def set_up(self):
        root = logging.getLogger()
        # 将的所有使用Logger模块生成的logger设置一样的logger level
        for log in root.manager.loggerDict.keys():
            root.getChild(log).setLevel(getattr(logging, self.level, 10))

        if self.stdout:
            self.set_handler(logging.StreamHandler(sys.stdout))
        elif self.log_file:
            os.makedirs(dir, exist_ok=True)
            file_handler = handlers.RotatingFileHandler(
                os.path.join(dir, "%s.log" % self.name),
                maxBytes=bytes,
                backupCount=self.backups)
            self.set_handler(file_handler)
        else:
            self.set_handler(UDPLogstashHandler(self.udp_host, self.udp_port))

    def set_handler(self, handler):
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(self._get_formatter())
        self.logger.addHandler(handler)
        self.debug("Logging to %s" % handler.__class__.__name__)

    def __getattr__(self, item):
        if item.upper() in logging._nameToLevel:
            func = getattr(self.logger, item)

            @wraps(func)
            def wrapper(*args, **kwargs):
                extra = kwargs.pop("extra", {})
                extra.setdefault("level", item)
                extra.setdefault("timestamp",
                                 datetime.datetime.utcnow().strftime(
                                     '%Y-%m-%dT%H:%M:%S.%fZ'))
                extra.setdefault("logger", self.name)
                extra.setdefault("threadname", current_thread().getName())
                kwargs["extra"] = extra
                return func(*args, **kwargs)
            return wrapper

        raise AttributeError

    def _get_formatter(self):
        if self.json:
            return JsonFormatter()
        else:
            return logging.Formatter(self.format_string)



from threading import RLock
from collections import UserDict
from collections.abc import MutableSet

from . import thread_safe_for_method


class ThreadSafeSet(MutableSet):
    def __init__(self, *args, **kwargs):
        self._data = set(*args, **kwargs)
        self.lock = RLock()

    @thread_safe_for_method
    def add(self, value):
        return self._data.add(value)

    @thread_safe_for_method
    def discard(self, value):
        return self._data.discard(value)

    @thread_safe_for_method
    def pop_all(self):
        while len(self._data):
            yield self.pop()

    @thread_safe_for_method
    def update(self, seq):
        self._data.update(seq)

    @thread_safe_for_method
    def __contains__(self, item):
        return item in self._data

    @thread_safe_for_method
    def __iter__(self):
        return iter(self._data)

    @thread_safe_for_method
    def __len__(self):
        return len(self._data)


class TreadSafeDict(UserDict):
    def __init__(self, *args, **kwargs):
        super(TreadSafeDict, self).__init__(*args, **kwargs)
        self.lock = RLock()

    @thread_safe_for_method
    def update(*args, **kwds):
        super(TreadSafeDict, args[0]).update(*args[1:], **kwds)

    @thread_safe_for_method
    def pop_all(self):
        while len(self):
            yield self.popitem()



import re
import os

from functools import partial
from argparse import ArgumentParser


def get_version(package):
    """
    Return package version as listed in `__version__` in `__init__.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    mth = re.search(r"__version__\s?=\s?['\"]([^'\"]+)['\"]", init_py)
    if mth:
        return mth.group(1)
    else:
        raise RuntimeError("Cannot find version!")


def install_requires():
    """
    Return requires in requirements.txt
    :return:
    """
    try:
        with open("requirements.txt") as f:
            return [line.strip() for line in f.readlines() if line.strip()]
    except OSError:
        return []


def change_version(index=None, dev=False, package_name=None):

    if not index:
        parser = ArgumentParser()
        parser.add_argument("-i", "--index", type=int, help="版本顺位", default=3)
        parser.add_argument("-d", "--dev", action="store_true", help="是否是开发模式")
        args = parser.parse_args()
        index = args.index
        dev = args.dev

    package = package_name or os.path.basename(os.path.abspath(os.getcwd())).replace("-", "_")
    with open(os.path.join(package, '__init__.py'), "r+", encoding="utf-8") as f:
        init_py = f.read()
        f.seek(0)
        buf = re.sub(
            r"(__version__\s?=\s?['\"])([^'\"]+)(['\"])",
            partial(_repl, index=int(index), dev=dev), init_py)
        f.write(buf)


def _repl(mth, index, dev):
    versions = mth.group(2).split(".")
    vs = versions[index - 1]
    length = len(vs)

    if vs.isdigit():
        new_vs = str(int(vs) + 1) + ("dev1" if dev else "")
    else:
        def _rep(mth, dev):
            string = mth.group(1)
            # 如果是版本开始，如：1d1中的第一个1，那么无论是否dev，1都可以保持不变。
            # 否则，则是第二个1或者d，当是dev=False时，这些都是要被舍弃的(正式版不含dev)
            # 只dev=True且是第二个1时，才需要自增1，如果是d，那么保持不变就可以。
            if mth.start():
                if dev:
                    if string.isdigit():
                        return str(int(string) + 1)
                else:
                    return ""

            return string

        new_vs = re.sub(r"(\d+|[a-zA-Z]+)",  partial(_rep, dev=dev), vs)
        if not new_vs[-1].isdigit():
            new_vs += "1"
    versions[index - 1] = new_vs
    blank = (length - len(new_vs)) * " "

    # 如果不是第三位+1。而是第二位或得第一位，则后续的位数应该清0
    for i in range(index, len(versions)):
        blank += " " * (len(versions[i]) - 1)
        versions[i] = "0"

    return mth.group(1) + ".".join(versions) + mth.group(3) + blank


import time
import traceback

from . import find_caller_name

__all__ = ["Blocker", "ExceptContext", "Timer"]


class Blocker(object):
    """
    有的时候我们需要将线程停止一段时间，通常我们选择调用time.sleep(..)，
    当我们需要sleep很长一段时间，比如一分钟以上时，如果这时我们选择关闭程序，
    而我们通过singal注册了关闭信号的监听器，用来改变当时程序的状态，
    如果置self.alive = False,由于time.sleep阻塞导致我们的程序当前线程无法获知alive状态，
    难以被关闭，通过使用Blocker，我们可以避免上述情况发生。
    eg:
    def myfunc(self):
        with Blocker(
        sleep_time, self, interval=1, notify=lambda x: not x.alive) as sm:
            if sm.is_notified:
                `我们知道是被唤醒了，而不是时间到了`
            ....
    """
    def __init__(self, block_time, instance, interval=1,
                 notify=lambda instance: False, immediately=False):
        """
        :param block_time: 需要阻塞的时长
        :param instance: 使用这个对象来动态告知manager什么时候要中断程序了，
        这个对象会被传递给notify回调函数
        :param interval: 如果这要中断程序，那么最长需要等待的时间间隔
        :param notify: 回调函数，接收instance，来决定是否唤醒阻塞。
        :param immediately: 是否不阻塞立即执行。
        """
        self.block_time = block_time
        self.instance = instance
        self.interval = interval
        self.notify = notify
        self.enter_time = time.time()
        self.immediately = immediately
        self.is_notified = True

    def __enter__(self):
        while not self.immediately and time.time() - \
                self.enter_time < self.block_time \
                and not self.notify(self.instance):
            time.sleep(self.interval)
        self.is_notified = self.notify(self.instance)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return exc_type is None


class ExceptContext(object):
    """
    异常捕获上下文
    eg:
    def test():
        with ExceptContext(Exception, errback=lambda name, *args:print(name)):
            raise Exception("test. ..")
    """
    def __init__(self, exception=Exception, func_name=None,
                 errback=lambda func_name, *args:
                 traceback.print_exception(*args) is None,
                 finalback=lambda got_err: got_err):
        """
        :param exception: 指定要监控的异常
        :param func_name: 可以选择提供当前所在函数的名称，回调函数会提交到函数，用于跟踪
        :param errback: 提供一个回调函数，如果发生了指定异常，
        就调用该函数，该函数的返回值为True时不会继续抛出异常
        :param finalback: finally要做的操作
        """
        self.errback = errback
        self.finalback = finalback
        self.exception = exception
        self.got_err = False
        self.err_info = None
        self.func_name = func_name or find_caller_name(is_func=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return_code = False
        if isinstance(exc_val, self.exception):
            self.got_err = True
            self.err_info = (exc_type, exc_val, exc_tb)
            return_code = self.errback(self.func_name, exc_type, exc_val, exc_tb)
        self.finalback(self.got_err)
        return return_code


class Timer(object):
    """
    计时器，对于需要计时的代码进行with操作：
    with Timer() as timer:
        ...
        ...
    print(timer.cost)
    ...
    """
    def __init__(self, start=None):
        self.start = start if start is not None else time.time()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop = time.time()
        self.cost = self.stop - self.start
        return exc_type is None